if (!com) var com = {};
if (!com.forward) com.forward = {};

com.forward.invisiblehand = {

  quiet: true,
  version: "3.8.28",
  ioService: Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService),
  ext: null, 

  getPrefs: function () {
    var prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    return prefService.getBranch("extensions.invisiblehand.");
  },
  
  getProperties: function() {
    return JSON.parse(com.forward.invisiblehand.getPrefs().getCharPref("config")) || {}
  },
  
  onDOMLoad: function (event) {
    var doc = event.originalTarget;
    if (doc.nodeName && doc.nodeName != "#document") return;

    var win = doc.defaultView;
    
    if (!doc instanceof HTMLDocument) return; // only documents
    if (win != win.top) return; //only top window.  
    if (win.frameElement) return; // skip iframes/frames  
    if (doc.location.href.match(/^chrome:\/\//)) return;

    com.forward.invisiblehand.log('onDOMLoad');
    
    if (!com.forward.invisiblehand.urlListenerAttached) {
      com.forward.invisiblehand.urlListenerAttached = true;
      com.forward.invisiblehand.log("url listener attached");
      com.forward.invisiblehand.firefox.attachUrlChangeListener();
    }
    com.forward.invisiblehand.onDOMRefreshed(event, false);
  },
    
  onTabClose: function() {
    com.forward.invisiblehand.ext = null;
  },

  onDOMRefreshed: function (event, pageIsNotReloaded) {
    var document = event.originalTarget;
    var address = document.location.href;
    if(address.match(/^about:/)) return;

    com.forward.invisiblehand.log('onDOMRefreshed: ' + address + ' (reload: ' + !!pageIsNotReloaded + ')');
    var browser = gBrowser.getBrowserForDocument(document);

    if (!browser || pageIsNotReloaded) {
      // com.forward.invisiblehand.log('onDOMRefreshed triggered following the change in url ' + address);
      var googleTab = browser || com.forward.invisiblehand.firefox.triggeredByTab(address);
      if (googleTab) {
        com.forward.invisiblehand.log('Found a corresponding google tab');
        var pageUpdateTimeout = pageIsNotReloaded ? 1000 : 0;
        com.forward.invisiblehand.log('Page update timeout is ' + pageUpdateTimeout);        
        setTimeout(function () {
          com.forward.invisiblehand.prepareAndRun(googleTab);
        }, pageUpdateTimeout);
      } else {
        // com.forward.invisiblehand.log('Corresponding google tab not found');
        return;
      }
    } else {
      com.forward.invisiblehand.log('Preparing to run');
      com.forward.invisiblehand.prepareAndRun(browser);
    }
  },

  prepareAndRun: function (browser) {
    var prefs = com.forward.invisiblehand.getPrefs();
    if (!prefs.prefHasUserValue("config")) {
      var firstConfigValue = prefs.prefHasUserValue("options") ? prefs.getCharPref("options") : "{}" //this is for updating users
      prefs.setCharPref("config", firstConfigValue);
    }
    try {
      com.forward.invisiblehand.config = JSON.parse(prefs.getCharPref("config"));
    } catch(e) {
      com.forward.invisiblehand.config = {};
    }
    var b = new com.forward.invisiblehand.browser(browser);
    this.ext = new com.forward.invisiblehand.extension(b);
    var options = {
      'products-search-engines-enabled':  com.forward.invisiblehand.getPrefs().getBoolPref('searchenginesenabled'),
      'products-ebay-enabled':            com.forward.invisiblehand.getPrefs().getBoolPref('ebayenabled'),
      'flights-enabled':                  com.forward.invisiblehand.getPrefs().getBoolPref('flightsenabled'),
      'hotels-enabled':                  com.forward.invisiblehand.getPrefs().getBoolPref('hotelsenabled'),
      'products-related-items-enabled':   com.forward.invisiblehand.getPrefs().getBoolPref('relateditemsenabled')
      };
    this.ext.run({extensionOptions: options, extensionProperties: b.getProperties()});
    
  },

  onLoad: function () {
    var appcontent = document.getElementById("appcontent");
    if (appcontent) {
      appcontent.addEventListener("DOMContentLoaded", com.forward.invisiblehand.onDOMLoad, true);
      com.forward.invisiblehand.log("on DOM load handler set.");
    } else {
      com.forward.invisiblehand.log("Element id=appcontent not found.");
    }
    
    gBrowser.tabContainer.addEventListener("TabClose", com.forward.invisiblehand.onTabClose, false);  
  },
  
  zeroPad: function (num, count) {
    var numZeropad = num + '';
    while (numZeropad.length < count) {
      numZeropad = "0" + numZeropad;
    }
    return numZeropad;
  },

  log: function (message) {
    if (this.quiet) return;
    var aConsoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
    aConsoleService.logStringMessage(message);
    var date = new Date();
    var time = date.getSeconds() + '.' + com.forward.invisiblehand.zeroPad(date.getMilliseconds(), 3) + ': '
    dump(time + message + "\n");
  },

  httpGetWithoutCookies: function (url, callback) {
    com.forward.invisiblehand.log("Channeled request initiated");
    var channel = this.ioService.newChannelFromURI(this.ioService.newURI(url, null, null));
    var listener = new this.streamListener(channel, callback);
    channel.notificationCallbacks = listener;
    channel.owner = new com.forward.invisiblehand.channelOwner();
    this.log(channel.owner.data);
    channel.asyncOpen(listener, null);
    com.forward.invisiblehand.log("ajaxGetRequst async opened");
  },

  httpRequest: function (url, options) {
    if (typeof options === "undefined") options = {};
    var method = options.method || "GET";

    var xhr = new XMLHttpRequest();
    if (options.callback) {
      xhr.onreadystatechange = function () {
        if (xhr.readyState == 4) options.callback.call(this, xhr.responseText);
      };
    }
    xhr.open(method, url, true);
    xhr.send(options.data);
  },
  
  getLanguage: function () {
    return 'en';
  }

};

com.forward.invisiblehand.volatileStorage = {};

com.forward.invisiblehand.browser = function (browser) {
  // this.browser = browser;
  this.doc = browser.contentDocument;
  this.window = browser.contentDocument.defaultView.wrappedJSObject;
}

com.forward.invisiblehand.browser.prototype = {
  
  saveProperty: function(key, value, callback) {
    var properties = this.getProperties();
    properties[key] = value;
    com.forward.invisiblehand.getPrefs().setCharPref("config", JSON.stringify(properties));
    callback(properties);
  },

  getProperties: function() {
	  return com.forward.invisiblehand.getProperties();
  },
  
  windowHeight: function() {
    return this.window.innerHeight;
  },
  
  getUrl: function(url, callback) {
    var self = this;
    if (com.forward.invisiblehand.common.utils.domainIsOnWhitelist(url)) {
      com.forward.invisiblehand.httpRequest(url, { callback: callback });  
    } else {
      com.forward.invisiblehand.log("WARNING! Tried to access domain not on the whitelist. " + url);
    }
  },
  
  report: function(url) {
    com.forward.invisiblehand.httpRequest(url, { method: 'POST' });
  },
  
  getInnerHtml: function(callback) {
    var self = this;
    setTimeout(function() {callback(self.doc.body.innerHTML)}, 300);
  },
  
  cache: function(key, value, callback) {
    if ((callback == undefined) && (typeof value == "function")) {
      callback = value;
      value = undefined;
    }
    if (!callback) callback = function() {};
    
    var timestamp = new Date().getTime();
    if (value == undefined) {
      var cached = this.volatileStorage()[key];
      if (cached && timestamp - cached.timestamp < 86400000) {
        callback(cached.value);
      } else {
        callback(undefined);  
      }
    } else {
      this.volatileStorage()[key] = {value:value, timestamp:timestamp};
      callback(value);
    }
  },  

  volatileStorage: function() {
	  return com.forward.invisiblehand.volatileStorage;
  },
  
  imageUrl: function(image_name) {
    return 'resource://invisiblehandstyles/images/' + image_name;
  },

  setZIndex: function (id, index) {
    this.doc.getElementById(id).style.setProperty('z-index', index, '');
  },

  currentUrl: function () {
    return this.doc.location.href;
  },

  waitForInnerHtml: function (callback) {
    var self = this;
    setTimeout(function () { 
      callback(self.doc.body ? self.doc.body.innerHTML : ''); 
    }, 200);
  },

  navigateToPage: function (url) {
    gBrowser.addTab(url);
  },

  extensionVersion: function () {
    return com.forward.invisiblehand.version;
  },

  browserType: function () {
    return 'firefox';
  },
  
  initialiseJQuery: function() {
    com.forward.invisiblehand.log("Initialising jQuery");
    if (this.window.jQuery.fn.jquery == '1.7.1') {
      com.forward.invisiblehand.log("jQuery 1.7.1 already present");
      return this.window.jQuery.noConflict(true);
    }

    var scriptLoader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);
    com.forward.invisiblehand.log("trying to load jquery and set it on the window object");
    scriptLoader.loadSubScript("chrome://canitbecheaper/content/jquery-1.7.1.min.js", { window: this.window});
    com.forward.invisiblehand.log("loaded the subscript. window.jQuery  " + this.window.jQuery);
    var jQ = this.window.jQuery;
    com.forward.invisiblehand.log('Loaded JQuery version: ' + jQ.fn.jquery);
    return jQ.noConflict(true);
  },
  
  executeJavaScriptInSandbox: function(js, context) {
    try {
      // Wrap in a function to maintain compatibility with other browsers
      var wrappedJS = "function executeJS() {" + js + "}; executeJS();"
          
      var sandbox = new Components.utils.Sandbox("http://www.getinvisiblehand.com");
      for (property in context) {
        sandbox[property] = context[property];
      }

      Components.utils.evalInSandbox(wrappedJS, sandbox);
    } catch (e) {
      com.forward.invisiblehand.log("WARN: Error executing Javascript");    
      com.forward.invisiblehand.log(e);
    }
  },

  executeJavaScript: function(js, context) {
    js(context.html, context["$ih"], context.callback, context.scrapeField);
  },

  postMessage: function(message, elementId) {
    if (elementId) {
      var iframe = this.doc.getElementById(elementId);
      if (iframe && iframe.contentWindow) {
        iframe.contentWindow.postMessage(message, '*');
      }
    } else {
      this.window.postMessage(message, '*');
    }
  }
};

com.forward.invisiblehand.firefox = {

  cookieListener: {
    observe: function (aSubject, aTopic, aData) {
      // cookies from scraping attempts should never be stored on the computer,
      // otherwise it'll be against T&C of the retailers that we scrape prices from
      // We shouldn't send them either, or the items we've scraped will appear in 'recently
      // viewed items'.
      if (aTopic != 'http-on-examine-response' && aTopic != 'http-on-modify-request')
        return;
      var ourResponse;
      var ourRequest;
      try {
        aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
        ourResponse = (aTopic == 'http-on-examine-response') && (aSubject.originalURI != undefined);
        ourRequest = (aTopic == 'http-on-modify-request') && (aSubject.URI != undefined);
      } catch (e) { }
      if (!(ourRequest || ourResponse)) {
        return;
      }
      var url = aSubject.originalURI.spec || aSubject.URI.spec;
      var ourChannel = false;
      var owner = aSubject.owner;
      if (owner) {
        try {
          owner.QueryInterface(Components.interfaces.nsISupportsString);
          ourChannel = owner.data == 'InvisibleHand Channel';
          com.forward.invisiblehand.log("Owner is " + owner.data);
        } catch (e) { }
      }
      if (!ourChannel) {
        // com.forward.invisiblehand.log("This channel is not ours: " + url);
        return;
      }
      com.forward.invisiblehand.log("Our channel is confirmed: " + url);
      if (aTopic == 'http-on-examine-response') {
        aSubject.setResponseHeader('Set-Cookie', '', false);
        com.forward.invisiblehand.log("Removed cookie for incoming " + aSubject.originalURI.spec);
      } else if (aTopic == 'http-on-modify-request') {
        aSubject.setRequestHeader('Cookie', '', false);
        com.forward.invisiblehand.log("Removed cookie for outgoing " + aSubject.URI.spec);
      }
    },

    QueryInterface: function (aIID) {
      if (aIID.equals(Components.interfaces.nsISupports) ||
          aIID.equals(Components.interfaces.nsIObserver))
        return this;
      throw Components.results.NS_NOINTERFACE;
    }
  },

  registerCookieObserver: function () {
    var observerService = this.getObserverService();
    observerService.addObserver(com.forward.invisiblehand.firefox.cookieListener, "http-on-examine-response", false);
    observerService.addObserver(com.forward.invisiblehand.firefox.cookieListener, "http-on-modify-request", false);
  },
  
  attachUrlChangeListener: function() {
    if (typeof gBrowser.addTabsProgressListener != 'function') { // for some reason it's not always immediately available
      com.forward.invisiblehand.log("gBrowser not ready yet...")
      setTimeout(com.forward.invisiblehand.firefox.attachUrlChangeListener, 100);
      return;
    }
    com.forward.invisiblehand.log("Attaching the progress listener");    
    gBrowser.addTabsProgressListener({
      lastUrl: '', // I don't get why onLocationChange is called multiple times under some circumstances :(
      QueryInterface: function (aIID)
          {
              if (aIID. equals (Components. interfaces. nsIWebProgressListener) ||
                  aIID. equals (Components. interfaces. nsISupportsWeakReference) ||
                  aIID. equals (Components. interfaces. nsIXULBrowserWindow) ||
                  aIID. equals (Components. interfaces. nsISupports))
                  return this;
              throw Components. results. NS_NOINTERFACE;
          },
      onLocationChange: function(aBrowser, webProgress, request, location) {
        if (this.lastUrl == location.spec) {
          com.forward.invisiblehand.log("Skipping location change since it has already been processed");
          return;
        }
        com.forward.invisiblehand.log("Tab location change: " + location.spec);
        this.lastUrl = location.spec;
        
        if (location.spec.match(/google/i)) {
          var doc = aBrowser.contentDocument;
          com.forward.invisiblehand.onDOMRefreshed({originalTarget: doc}, true);
        }           
      },
      onProgressChange: function(aBrowser, aWebProgress, aRequest, curSelf, maxSelf, curTot, maxTot) {},
      onStateChange: function(aBrowser, aWebProgress, aRequest, aFlag, aStatus) {},
      onStatusChange: function(aBrowser, aWebProgress, aRequest, aStatus, aMessage) {},
      onSecurityChange: function(aBrowser, aWebProgress, aRequest, aState) {},
      onRefreshAttempted: function(aBrowser, aWebProgress, aRefreshURI, aMillis,aSameURI) {
              // must return true. or the refresh action will be blocked
              return true;
      },
      onLinkIconAvailable: function(aBrowser) {} 
    }, Ci.nsIWebProgress.NOTIFY_LOCATION);
  },  
  
  registerUninstallObserver: function() {
    com.forward.invisiblehand.firefox.registerUninstallObserverForFF3();
    com.forward.invisiblehand.firefox.registerUninstallObserverForFF4();
  },
  
  registerUninstallObserverForFF4: function() {
    try {
      Components.utils.import("resource://gre/modules/AddonManager.jsm");      
    } catch(e) {
      com.forward.invisiblehand.log("Could not get an addon manager");
      return;
    }
    var listener = {
      onDisabling: function(addon) {
        com.forward.invisiblehand.log("Disabling!");
      },
      onUninstalling: function(addon, needsRestart) {
        com.forward.invisiblehand.log("Uninstallation of addon " + addon.id);        
        if (addon.id != 'canitbecheaper@trafficbroker.co.uk') return;
        com.forward.invisiblehand.firefox.askWhyUninstalling();
      }
    }
    AddonManager.addAddonListener(listener);
    com.forward.invisiblehand.log("registered uninstall listener");    
  },
  
  askWhyUninstalling: function() {
    var window_params = {inn:{}, out:null};       
    window.openDialog("chrome://canitbecheaper/content/uninstall_dialog.xul", "",
      "chrome, dialog, modal, resizable=yes", window_params).focus();
    if (window_params.out) {
      // alert("Thank you for your feedback. We're continuously working to make this tool better.")
      // User clicked ok
    } else {
      // User clicked cancel
    }
    var description = window_params.out.description ? escape(window_params.out.description) : '';
    var reason = window_params.out.reason;
    var params = 'action=uninstallation&uid=' + com.forward.invisiblehand.getProperties().uid + '&locale=' + com.forward.invisiblehand.getLanguage() + '&reason=' + reason + '&description=' + description;
    com.forward.invisiblehand.log(params);
    com.forward.invisiblehand.httpRequest('http://statistics.invisiblehand.co.uk?' + params, {});
    gBrowser.addTab('http://productsiframe.invisiblehand.co.uk/uninstall');
  },
  
  registerUninstallObserverForFF3: function () {
    var uninstallObserver = {
      observe: function (subject, topic, data) {
        if (topic == "em-action-requested") {
          subject.QueryInterface(Components.interfaces.nsIUpdateItem);
          com.forward.invisiblehand.log("Action observed: " + subject.id);
          if (subject.id == 'canitbecheaper@trafficbroker.co.uk') {
            if (data == "item-uninstalled") {
              com.forward.invisiblehand.firefox.askWhyUninstalling();
            }
          }
        }
      },
      register: function () {
        com.forward.invisiblehand.firefox.getObserverService().addObserver(this, "em-action-requested", false);
      },
      unregister: function () {
        com.forward.invisiblehand.firefox.getObserverService().removeObserver(this, "em-action-requested");
      }
    };
    uninstallObserver.register();
  },

  getObserverService: function () {
    return Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  },

  triggeredByTab: function (address) {
    for (var i = gBrowser.browsers.length - 1; i >= 0; i--) {
      var browser = gBrowser.getBrowserAtIndex(i);
      if (browser.currentURI.spec == address) {
        return browser;
      }
    }
    return false;
  },
  
};

window.addEventListener("load", com.forward.invisiblehand.onLoad, false);
com.forward.invisiblehand.firefox.registerUninstallObserver();
com.forward.invisiblehand.firefox.registerCookieObserver();
