com.forward.invisiblehand.constants = {
    statServer: 'data.invisiblehand.co.uk',
    infoUrl: "http://www.getinvisiblehand.com/help/gettingstarted",
    geoEndpoint: "http://ihgeoip.invisiblehand.co.uk/v1/TrnJep8HTPbfbHsGDaJ7r3Fdxt8wowDeBGkZDuJaEvPrU914xPSfJg1LANm8Xt/",
    mixpanelMainToken: "a69578d70ac0dd9da90935597f537fca"
};

com.forward.invisiblehand.uuid = (function() {
  /*
  * Generate a RFC4122(v4) UUID
  *
  * Documentation at https://github.com/broofa/node-uuid
  */

  var BufferClass = typeof(Buffer) == 'function' ? Buffer : Array;

  var _buf = new BufferClass(16);

  var toString = [];
  var toNumber = {};
  for (var i = 0; i < 256; i++) {
    toString[i] = (i + 0x100).toString(16).substr(1);
    toNumber[toString[i]] = i;
  }

  function parse(s) {
    var buf = new BufferClass(16);
    var i = 0;
    s.toLowerCase().replace(/[0-9a-f][0-9a-f]/g, function(octet) {
      buf[i++] = toNumber[octet];
    });
    return buf;
  }

  function unparse(buf) {
    var tos = toString, b = buf;
    return tos[b[0]] + tos[b[1]] + tos[b[2]] + tos[b[3]] + '-' +
           tos[b[4]] + tos[b[5]] + '-' +
           tos[b[6]] + tos[b[7]] + '-' +
           tos[b[8]] + tos[b[9]] + '-' +
           tos[b[10]] + tos[b[11]] + tos[b[12]] +
           tos[b[13]] + tos[b[14]] + tos[b[15]];
  }

  var ff = 0xff;

  var useCrypto = this.crypto && crypto.getRandomValues;
  var rnds = useCrypto ? new Uint32Array(4) : new Array(4);

  function uuid(fmt, buf, offset) {
    var b = fmt != 'binary' ? _buf : (buf ? buf : new BufferClass(16));
    var i = buf && offset || 0;

    if (useCrypto) {
      crypto.getRandomValues(rnds);
    } else {
      rnds[0] = Math.random()*0x100000000;
      rnds[1] = Math.random()*0x100000000;
      rnds[2] = Math.random()*0x100000000;
      rnds[3] = Math.random()*0x100000000;
    }

    var r = rnds[0];
    b[i++] = r & ff;
    b[i++] = r>>>8 & ff;
    b[i++] = r>>>16 & ff;
    b[i++] = r>>>24 & ff;
    r = rnds[1];
    b[i++] = r & ff;
    b[i++] = r>>>8 & ff;
    b[i++] = r>>>16 & 0x0f | 0x40; // See RFC4122 sect. 4.1.3
    b[i++] = r>>>24 & ff;
    r = rnds[2];
    b[i++] = r & 0x3f | 0x80; // See RFC4122 sect. 4.4
    b[i++] = r>>>8 & ff;
    b[i++] = r>>>16 & ff;
    b[i++] = r>>>24 & ff;
    r = rnds[3];
    b[i++] = r & ff;
    b[i++] = r>>>8 & ff;
    b[i++] = r>>>16 & ff;
    b[i++] = r>>>24 & ff;

    return fmt === undefined ? unparse(b) : b;
  }

  uuid.parse = parse;
  uuid.unparse = unparse;
  uuid.BufferClass = BufferClass;

  return uuid;
})();
com.forward.invisiblehand.common = {}

com.forward.invisiblehand.common.utils = {

    buildUrl: function (server, params, secure) {
      var parts = [secure ? 'https' : 'http', '://', server];
      if (params) parts.push("?");

      for (var i in params) {
        if (typeof params[i] !== "undefined" && params[i] !== null) {
          parts.push('&');
          parts.push(i);
          parts.push('=');
          parts.push(encodeURIComponent(params[i]));
        }
      }

      return parts.join('');
    },

    getRegionFrom: function(countryCode) {
      var region = countryCode === 'GB' ? 'UK' : countryCode
      return region;
    },

    extend: function(destination, source) {
      for (var property in source) {
          if (source.hasOwnProperty(property)) {
              destination[property] = source[property];
          }
      }
      return destination;
    },

    domainIsOnWhitelist: function(url) {
      var match = url.match(/^https?:\/\/(?:[a-z0-9]+\.)*([a-z0-9\-]+\.(co\.uk|com|de|ca|net))/)
      if (!match) {
        com.forward.invisiblehand.log("Could not extract domain from url : " + url);
        return false;
      }
      var domain = match[1];

      for (i in com.forward.invisiblehand.domainWhitelist) {
        if (domain == com.forward.invisiblehand.domainWhitelist[i]) {
          com.forward.invisiblehand.log("Domain " + domain + " is on whitelist");
          return true;
        }
      }

      com.forward.invisiblehand.log("Domain " + domain + " is NOT on whitelist");
      return false;
    }
};

(function() {
  var Messaging;

  Messaging = (function() {

    Messaging.name = 'Messaging';

    function Messaging(browser) {
      this.browser = browser;
      this.doc = this.browser.doc;
      this.window = this.browser.window;
      this.messageHandlers = {};
      this.addEventListener();
    }

    Messaging.prototype.addEventListener = function() {
      var func,
        _this = this;
      func = function(evt) {
        return _this.receiveMessage(evt);
      };
      if (this.window.addEventListener) {
        return this.window.addEventListener("message", func, false);
      } else if (this.window.attachEvent) {
        return this.window.attachEvent("onmessage", func);
      }
    };

    Messaging.prototype.setMessageHandler = function(messageType, func) {
      return this.messageHandlers[messageType] = func;
    };

    Messaging.prototype.setMessageHandlers = function(handlers) {
      var func, messageType, _results;
      _results = [];
      for (messageType in handlers) {
        func = handlers[messageType];
        _results.push(this.messageHandlers[messageType] = func);
      }
      return _results;
    };

    Messaging.prototype.receiveMessage = function(evt) {
      var messageHandler, payload;
      com.forward.invisiblehand.log("receiveMessage: " + evt.data);
      payload = {};
      try {
        payload = JSON.parse(evt.data);
      } catch (e) {
        return;
      }
      messageHandler = this.messageHandlers[payload.type || payload.message];
      if (messageHandler) {
        com.forward.invisiblehand.log("Got handler for " + payload.type);
        return messageHandler(payload);
      } else {
        return com.forward.invisiblehand.log("Don't know what to do with: " + payload);
      }
    };

    Messaging.prototype.sendMessage = function(message, elementId) {
      var j = JSON.stringify(message);
      com.forward.invisiblehand.log("sendMessage: " + j + ", elementId: " + elementId);
      return this.postMessage(j, elementId);
    };

    Messaging.prototype.postMessage = function(message, elementId) {
      return this.browser.postMessage(message, elementId);
    };

    return Messaging;

  })();

  com.forward.invisiblehand.Messaging = Messaging;

}).call(this);
(function() {
  var Iframe, Stylesheet,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  Iframe = (function() {

    Iframe.name = 'Iframe';

    function Iframe(htmlId, url, extension, currentPage) {
      this.htmlId = htmlId;
      this.url = url;
      this.extension = extension;
      this.currentPage = currentPage;
      this.removeElement = __bind(this.removeElement, this);

      this.remove = __bind(this.remove, this);

      this.show = __bind(this.show, this);

      this.collapse = __bind(this.collapse, this);

      this.expand = __bind(this.expand, this);

      this.ready = __bind(this.ready, this);

      this.originalBodyMarginTop = 0;
      this.messaging = this.extension.messaging;
      this.messageQueue = [];
      this.isReady = false;
      this.doc = this.extension.browser.doc;
      this.stylesheetId = 'invisiblehand-overriding-stylesheet';
      this.shown = false;
      this.messaging.setMessageHandlers({
        closeNotification: this.remove,
        expandIframe: this.expand,
        iframeReady: this.ready,
        showNotification: this.show,
        collapseIframe: this.collapse
      });
    }

    Iframe.prototype.notify = function(type, msg) {
      var message;
      message = msg;
      message.type = type;
      if (this.isReady) {
        return this.messaging.sendMessage(message, this.htmlId);
      } else {
        return this.messageQueue.push(message);
      }
    };

    Iframe.prototype.ready = function() {
      var message, _i, _len, _ref, _results;
      com.forward.invisiblehand.log("ready");
      this.isReady = true;
      _ref = this.messageQueue;
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        message = _ref[_i];
        _results.push(this.messaging.sendMessage(message, this.htmlId));
      }
      return _results;
    };

    Iframe.prototype.expand = function() {
      com.forward.invisiblehand.log("Expand iframe");
      return this.iframeElement.style.height = '100%';
    };

    Iframe.prototype.collapse = function(message) {
      com.forward.invisiblehand.log("Collapse iframe");
      this.show(message);
      return this.iframeElement.style.height = this.height + "px";
    };

    Iframe.prototype.show = function(message) {
      if (!this.shown) {
        com.forward.invisiblehand.log("Show iframe");
        this.height = message.notificationHeight || this.height;
        this.iframeElement.style.height = this.height + "px";
        this.doc.getElementsByTagName("body")[0].setAttribute('style', "margin-top: " + (this.height + this.originalBodyMarginTop) + "px !important;");
        this.applyStylesheet();
        return this.shown = true;
      }
    };

    Iframe.prototype.applyStylesheet = function() {
      if (this.currentPage.retailer.style_override) {
        this.stylesheet = new Stylesheet(this.stylesheetId, this.doc, this.currentPage.retailer.style_override);
        return this.stylesheet.insert(this.height);
      }
    };

    Iframe.prototype.remove = function() {
      return this.removeElement(this.iframeElement);
    };

    Iframe.prototype.removeElement = function(element) {
      var remove,
        _this = this;
      com.forward.invisiblehand.log("Remove iframe");
      element.style.height = "0px";
      this.doc.getElementsByTagName("body")[0].setAttribute('style', "margin-top: " + this.originalBodyMarginTop + "px;");
      if (this.stylesheet) {
        this.stylesheet.remove();
      }
      remove = function() {
        return _this.doc.body.removeChild(element);
      };
      return setTimeout(remove, 1000);
    };

    Iframe.prototype.insert = function() {
      var oldElement;
      com.forward.invisiblehand.log("Insert iframe");
      if (oldElement = this.doc.getElementById(this.htmlId)) {
        com.forward.invisiblehand.log("Removing old iframe");
        this.removeElement(oldElement);
      } else if (this.doc.body.style.marginTop) {
        this.originalBodyMarginTop = parseInt(this.doc.body.style.marginTop.replace(/[^\d]/g, "")) || 0;
      }
      com.forward.invisiblehand.log("originalBodyMarginTop: " + this.originalBodyMarginTop);
      this.iframeElement = this.doc.createElement("iframe");
      this.iframeElement.id = this.htmlId;
      this.iframeElement.src = this.url;
      this.iframeElement.scrolling = 'no';
      this.iframeElement.frameBorder = '0';
      this.iframeElement.width = '100%';
      this.iframeElement.height = '0px';
      this.iframeElement.style.position = 'fixed';
      this.iframeElement.style.zIndex = '10000001';
      this.iframeElement.style.left = 0;
      this.iframeElement.style.right = 0;
      this.iframeElement.style.top = 0;
      this.iframeElement.style.width = '100%';
      return this.doc.body.insertBefore(this.iframeElement, this.doc.body.firstChild);
    };

    return Iframe;

  })();

  com.forward.invisiblehand.Iframe = Iframe;

  Stylesheet = (function() {

    Stylesheet.name = 'Stylesheet';

    function Stylesheet(elementId, doc, stylesheet) {
      this.elementId = elementId;
      this.doc = doc;
      this.stylesheet = stylesheet;
    }

    Stylesheet.prototype.insert = function(height) {
      var existingStylesheet, head, newCss, stylesheetElement;
      newCss = this.process(height);
      stylesheetElement = this.element(newCss);
      head = this.doc.getElementsByTagName("head").item(0);
      existingStylesheet = this.doc.getElementById(this.elementId);
      if (existingStylesheet) {
        head.removeChild(existingStylesheet);
      }
      return head.appendChild(stylesheetElement);
    };

    Stylesheet.prototype.remove = function() {
      var existingStylesheet, head;
      existingStylesheet = this.doc.getElementById(this.elementId);
      head = this.doc.getElementsByTagName("head").item(0);
      if (existingStylesheet) {
        return head.removeChild(this.doc.getElementById(this.elementId));
      }
    };

    Stylesheet.prototype.process = function(height) {
      var match, newCss, newHeight, regex;
      regex = new RegExp(/<<([-\d]+)>>/g);
      newCss = this.stylesheet;
      while ((match = regex.exec(this.stylesheet))) {
        newHeight = parseInt(match[1]) + height;
        newCss = newCss.replace(match[0], newHeight);
      }
      return newCss;
    };

    Stylesheet.prototype.element = function(stylesheet) {
      var styleElement;
      styleElement = this.doc.createElement("style");
      styleElement.type = 'text/css';
      styleElement.id = this.elementId;
      styleElement.appendChild(this.doc.createTextNode(stylesheet));
      return styleElement;
    };

    return Stylesheet;

  })();

}).call(this);
com.forward.invisiblehand.common.htmlScraper = function (data, scraper, browser) {
  this.data    = data;
  this.scraper = scraper;
  this.results = {};
  this.fields  = this.extractFields();
  this.getUrl  = browser.getUrl;
  this.browser = browser;
}

com.forward.invisiblehand.common.htmlScraper.prototype = {

  scrape: function(externalCallback) {
    if (this.scraper.js) return this.jsScrape(this.scraper.js, externalCallback);
    this.scrapeAllFields(externalCallback);
  },

  extractFields: function() {
    var fields = [];
    for (var fieldName in this.scraper)
      fields.push(fieldName);
    return fields;
  },

  scrapeAllFields: function(externalCallback) {
    var field = this.fields.pop();
    if (!field) return externalCallback(this.results);
    var self = this;
    this.scrapeField(field, function(){
      self.scrapeAllFields(externalCallback);
    });
  },

  normalizePrice: function(price) {
    price = price.replace(/[^,.\d]/, '');
    if (this.isGermanPrice(price)) return this.normalizeGermanPrice(price);
    if (this.isNormalPrice(price)) return this.normalizeNormalPrice(price);
    return price;
  },

  isGermanPrice: function(price) {
    return (price.indexOf(',') > -1) && (price.indexOf(',') > price.indexOf('.'));
  },

  isNormalPrice: function(price) {
    return (price.indexOf('.') > -1) && (price.indexOf('.') > price.indexOf(','));
  },

  normalizeGermanPrice: function(price) { // 1.800,99
    price = price.replace('.', '').replace(',', '.');
    return parseFloat(price);
  },

  normalizeNormalPrice: function(price) { // 1,800.99
    price = price.replace(',', '');
    return parseFloat(price);
  },

  normalizeScrapedValue: function(field, scrapedValue) {
    if (!scrapedValue) return;
    scrapedValue = scrapedValue.toString().replace(/[\x00-\x1F"]/g, ''); // replaces control characters and a quote to avoid problems with JSON later
    if (field == 'price') return this.normalizePrice(scrapedValue) || undefined;
    return scrapedValue;
  },

  scrapeField: function(field, callback){
		var self = this;
    this.tryScrapingUsingRegex(field, function() {
			if (self.results[field])
				callback()
			else
				self.tryScrapingUsingJS(field, callback);
		});
  },

  tryScrapingUsingRegex: function(field, callback) {
    if (!this.scraper[field].regex) {
			callback();
			return;
		}
    var scraped = this.regexScrape(this.data, this.scraper[field].regex);
    return this.processScrapedValue(field, scraped, callback);
  },

  tryScrapingUsingJS: function(field, callback) {
    if (!this.scraper[field].js) {
      callback();
      return;
    }
    var self = this;
    this.jsScrape(this.scraper[field].js, function(scrapedValue) {
      self.processScrapedValue(field, scrapedValue, callback);
    });
  },

  processScrapedValue: function(field, scrapedValue, callback) {
    if(!scrapedValue) {
      callback();
      return;
    }
    var scrapedValue = this.normalizeScrapedValue(field, scrapedValue);
    this.results[field] = scrapedValue;
    callback();
    return !!scrapedValue;
  },

  regexScrape: function(html, regex) {
    var match = html.match(new RegExp(regex, 'i'));
    if (match) return match[1];
    return;
  },

  jsScrape: function(js, externalCallback) {
    this.browser.executeJavaScriptInSandbox(js, { html: this.data, getHttp: this.getUrl, callback: externalCallback, scrapeField: this.regexScrape });
  }

};

com.forward.invisiblehand.common.domScraper = function (browser, scraper) {
  this.browser = browser;
  this.$       = browser.initialiseJQuery();
  this.address = browser.currentUrl();
  this.scraper = scraper;
  this.results = {}; // we want to keep them since we need to return everything for delayed and volatile scrapers
}

com.forward.invisiblehand.common.domScraper.prototype = {

  scrape: function(externalCallback) {
    var self = this;
    this.browser.getInnerHtml(function(html) {
      self.html = html;
      self.scrapeAllFields(externalCallback);
    });

  },

  /*
  * Most scrapers are syncronous, that's why we are using the for loop and firing the callback before returning
  * If we encounter an asynchronous scraper, the callback will be called one or more times
  * If there are no synchronous scrapers provided, the callback will not be fired immediately
  */
  scrapeAllFields: function(externalCallback) {
    if (this.scraper.js) {
      this.tryScrapingUsingJsAsync(this.scraper.js, externalCallback);
    }
    for (var fieldName in this.scraper) {
      var scrapedValue = this.scrapeField(fieldName, externalCallback);
      if (scrapedValue) {
        this.results[fieldName] = scrapedValue;
      }
    }
    externalCallback(this.results);
  },

  /*
  * If it's possible to scrape syncronously, we just return the value
  * Otherwise we should return undefined and call the externalCallback when the data is ready
  */
  scrapeField: function(field, externalCallback) {
    var timeout = parseInt(this.scraper[field].wait_for || this.scraper[field].rescrape_for); // it's in seconds on the server-side

    return timeout ? this.scrapeChangingField(field, timeout, externalCallback) : this.tryAllScrapingTechniques(field);
  },

  tryAllScrapingTechniques: function(field) {
    var scraper = this.scraper[field];
    return this.tryScrapingUsingRegex(scraper.regex)
        || this.tryScrapingUsingJsSync(scraper.js)
        || this.tryScrapingUrl(scraper.url_param)
        || this.tryScrapingElementContent(scraper.element_text || scraper.form_field)
        || this.tryScrapingMinValue(scraper.min_value)
        || this.tryScrapingSumOfValues(scraper.sum_values);
  },

  tryScrapingSumOfValues: function(selector) {
    if (!selector) return;
    var self = this;
    var sum = 0;

    this.$(selector).each(function(index, element) {
      var floatValue = self.parseFloatAggressively($(element).text());
      if (floatValue) sum += floatValue;
    });

    return Math.round(sum * 100) / 100.0;
  },

  parseFloatAggressively: function(text) {
    return parseFloat(text.replace(/[^\d\.]/g, ""));
  },

  tryScrapingMinValue: function(selector) {
    if (!selector) return;
    var self = this;
    var values = [];

    this.$(selector).each(function(index, element) {
      var floatValue = self.parseFloatAggressively($(element).text());
      if (floatValue) values.push(floatValue);
    });

    if (values.length > 0) return Math.min.apply(Math, values);
  },

  tryScrapingElementContent: function(selector) {
    if (!selector) return;
    var elements = this.$(selector);
    if (elements.length > 0) return elements.eq(0).val() || elements.eq(0).text();
  },

  tryScrapingUrl: function(regex) {
    if (!regex) return;
    var unescaped = unescape(this.address);
    var match = unescaped.match(new RegExp(regex + "=(.+?)&", 'i'));
    if (match) return match[1];
  },

  scrapeChangingField: function(field, timeout, externalCallback) {
    var intervalId,
        intervalInvocations = 0,
        self = this,
        stopWhenScraped = !!this.scraper[field].wait_for;

    intervalId = setInterval(function() {
      if (++intervalInvocations >= timeout) clearInterval(intervalId); // it's ++intervalInvocations because we have to wait 1 sec before the first invocation
      self.results[field] = self.tryAllScrapingTechniques(field);
      if (self.results[field]) {
        if (stopWhenScraped) clearInterval(intervalId);
        externalCallback(self.results);
      }
    }, 1000); // this has to be 1 second since it's specified in seconds on the server-side
  },

  tryScrapingUsingRegex: function(regex) {
    return this.scrapeHtmlUsingRegex(this.html, regex);
  },

  scrapeHtmlUsingRegex: function(html, regex) {
    if (!regex) return;
    html = html.replace(/\ssizcache="\d+"/, "").replace(/\ssizset="\d+"/, "");
    var match = html.match(new RegExp(regex, 'i'));
    if (match) return match[1];
  },

  tryScrapingUsingJsSync: function(jsCode) {
    if (!jsCode) return;

    var results;
    this.jsScrape(jsCode, function(scrape) { results = scrape });
    return results;
  },

  tryScrapingUsingJsAsync: function(jsCode, externalCallback) {
    if (!jsCode) return;
    var self = this;
    this.jsScrape(jsCode, function(scrape) {
      for (field in scrape) {
        self.results[field] = scrape[field];
      }
      externalCallback(self.results);
    });
  },

  jsScrape: function(js, externalCallback) {
    this.browser.executeJavaScript(js, { html: this.html, "$ih": this.$, callback: externalCallback, scrapeField: this.scrapeHtmlUsingRegex });
  }


};

com.forward.invisiblehand.extension = function (browser) {
  this.browser = browser;
  this.properties = {};
  this.notificationsCount = {};
  this.trackingDivId = "tc_container";
  this.executionCancelled = null;
};

com.forward.invisiblehand.extension.prototype = {

    setProperties: function (properties) {
        properties.domain = "invisiblehand.co.uk";
        properties.trackingCode = "ih";
        if(!properties.source || !properties.source.match(/^conduit/)) properties.source = "ih";
        this.properties = properties;
    },

    setProperty: function (key, value) {
        var self = this;
        this.browser.saveProperty(key, value, function (props) {
            self.setProperties(props);
        });
    },

    htmlScraper: function (data, scraper) {
        return new com.forward.invisiblehand.common.htmlScraper(data, scraper, this.browser);
    },

    domScraper: function (scraper) {
        return new com.forward.invisiblehand.common.domScraper(this.browser, scraper);
    },

    getGeoData: function(callback){
      this.browser.getUrl(com.forward.invisiblehand.constants.geoEndpoint, function(raw) {
        var data = JSON.parse(raw);
        callback({
          region: com.forward.invisiblehand.common.utils.getRegionFrom(data.countryCode),
          ip: data.ipAddress});
      });
    },

    installationSource: function () {
        return this.properties.source;
    },

    hideNotification: function () {
        this.browser.removeIframe();
    },

    isFirstRun: function () {
        return !this.properties.hasRun;
    },

    generateUid: function () {
        return com.forward.invisiblehand.uuid();
    },

    generateUidForCurrentUser: function () {
        return "c" + this.generateUid();
    },

    onFirstRun: function () {
        com.forward.invisiblehand.log('First Run');
        var self = this;

        self.setProperty("hasRun", true);

        var uid = this.properties.uid || self.generateUid();
        this.properties.uid = uid; // So not to be rewritten onFirstRunToday
        self.setProperty("uid", uid);

        self.browser.navigateToPage(com.forward.invisiblehand.constants.infoUrl);

        var params = {
            'extension': self.browser.extensionVersion(),
            'browser': self.browser.browserType(),
            'src': self.installationSource(),
            'uid': uid
        };

        this.trackByUser('installation', uid, params);

        var url = com.forward.invisiblehand.common.utils.buildUrl(com.forward.invisiblehand.constants.statServer + '/installation', params);
        com.forward.invisiblehand.httpRequest(url);
    },

    isFirstRunToday: function () {
        var lastUsed = this.properties.activeUser;
        var d = new Date();
        var today = "" + d.getFullYear() + "-" + (d.getMonth() + 1) + "-" + d.getDate();
        if (lastUsed != today) {
            this.setProperty("activeUser", today);
            return true;
        }

        return false;
    },

    trackByUser: function(event, uid, params){
      this._trackOnMixpanel(event,com.forward.invisiblehand.constants.mixpanelMainToken, uid, params);
    },

     _trackOnMixpanel: function (event, token, uniqueId, params){
      var self = this;
      this.getGeoData(function(geo){
        var mixpanelTracking = { "event": event,
          "properties": {
            'extension': self.browser.extensionVersion(),
            'browser': self.browser.browserType(),
            'src': self.properties.source,
            'uid': params["uid"],
            'token': token,
            'ip': geo.ip,
            'region': geo.region,
            'distinct_id' : uniqueId
          }
        }

        mixpanelTracking['properties'] = com.forward.invisiblehand.common.utils.extend(mixpanelTracking['properties'], params);

        com.forward.invisiblehand.log('Tracking on Mixpanel:' + JSON.stringify(mixpanelTracking));

        var encoded = window.btoa(unescape(encodeURIComponent(JSON.stringify(mixpanelTracking))));
        com.forward.invisiblehand.httpRequest("http://api.mixpanel.com/track/?data=" + encoded, { callback: function(response){
          com.forward.invisiblehand.log("Response from mixpanel: " + response); }} );
      });
    },

    onFirstRunToday: function () {
        com.forward.invisiblehand.log('First Run Today');

        var uid = this.properties.uid;
        if (!uid) {
            com.forward.invisiblehand.log("No UID for current user");
            uid = this.generateUidForCurrentUser();
            this.setProperty("uid", uid);
        }

        var params = {
            'extension': this.browser.extensionVersion(),
            'browser': this.browser.browserType(),
            'src': this.properties.source,
            'uid': uid
        };

        this.trackByUser('active', uid, params);

        var url = com.forward.invisiblehand.common.utils.buildUrl(com.forward.invisiblehand.constants.statServer + '/active', params);
        com.forward.invisiblehand.httpRequest(url);

        com.forward.invisiblehand.trackActiveUser && com.forward.invisiblehand.trackActiveUser();
    },

    setupVerticals: function () {
      var self = this;
      self.messaging = new com.forward.invisiblehand.Messaging(self.browser);

      self.products = new com.forward.invisiblehand.products(self, self.options);
      self.hotels = new com.forward.invisiblehand.hotels(self, self.options);
      self.flights = new com.forward.invisiblehand.flights(self, self.options);
      self.rentals = new com.forward.invisiblehand.rentals(self, self.options);
      self.tickets = new com.forward.invisiblehand.tickets(self, self.options);

      self.pageIntegration = new com.forward.invisiblehand.PageIntegration(self);
    },

    runVerticals: function () {
      this.products.run();
      this.hotels.run();
      this.flights.run();
      this.rentals.run();
      this.tickets.run();
    },

    trackBrokenLocalStorage: function () {
    },

    testLocalStorage: function (worksCallback, doesntWorkCallback) {
        var testKey = "localstoragetest";
        var testValue = this.generateUid();
        var self = this;
        this.browser.saveProperty(testKey, testValue, function (props) {
            if (props[testKey] == testValue) {
                worksCallback();
            } else {
                self.trackBrokenLocalStorage();
                doesntWorkCallback();
            }
        });
    },

    runSetup: function (callback) {
        if (this.isFirstRun()) {
            this.onFirstRun();
        }
        if (this.isFirstRunToday()) {
            this.onFirstRunToday();
        }
        callback();
    },

    run: function (params) {
      com.forward.invisiblehand.log('Starting extension.');

      this.setProperties(params.extensionProperties);
      this.options = params.extensionOptions;

      this.setupVerticals();

      var self = this;
      this.testLocalStorage(
        function () { self.runSetup(function () { self.runVerticals(params); }) },
        function () { self.runVerticals(params) }
      );
    },

    refresh: function () {
      com.forward.invisiblehand.log('Refreshing extension.');
      this.runVerticals();
    }

};
(function() {
  var PageIntegration,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  PageIntegration = (function() {

    PageIntegration.name = 'PageIntegration';

    function PageIntegration(extension) {
      this.extension = extension;
      this.productScrape = __bind(this.productScrape, this);

      this.messaging = this.extension.messaging;
      this.messaging.setMessageHandlers({
        "ih.pageIntegration.productScrape": this.productScrape
      });
    }

    PageIntegration.prototype.productScrape = function(message) {
      var retailer,
        _this = this;
      com.forward.invisiblehand.log(this.extension.products.retailers);
      retailer = this.extension.products.retailers.findRetailer(message.url);
      com.forward.invisiblehand.log(retailer);
      if (!retailer) {
        this.messaging.sendMessage({
          type: "scrapeResult",
          url: message.url,
          scrape: {}
        });
      }
      com.forward.invisiblehand.log("Scraping " + message.url);
      return new com.forward.invisiblehand.RemoteScrape(this.extension).scrape(message.url, retailer, function(scrape) {
        return _this.messaging.sendMessage({
          type: "scrapeResult",
          url: message.url,
          scrape: scrape
        });
      });
    };

    return PageIntegration;

  })();

  com.forward.invisiblehand.PageIntegration = PageIntegration;

}).call(this);
(function() {
  var RemoteScrape;

  RemoteScrape = (function() {

    RemoteScrape.name = 'RemoteScrape';

    function RemoteScrape(extension) {
      this.extension = extension;
    }

    RemoteScrape.prototype.scrape = function(url, retailer, callback) {
      var _this = this;
      return this.extension.browser.getUrl(url, function(html) {
        var scraper;
        scraper = _this.extension.htmlScraper(html, retailer.scraper);
        return scraper.scrape(function(result) {
          return callback(result);
        });
      });
    };

    return RemoteScrape;

  })();

  com.forward.invisiblehand.RemoteScrape = RemoteScrape;

}).call(this);
com.forward.invisiblehand.currentPage = function (retailers, options, browser, extension) {
  this.retailer = retailers.getRetailer();
  this.url = this.fixUrlForPlay(browser.currentUrl());
  this.retailers = retailers;
  this.options = options;
  this.browser = browser;
  this.extension = extension;
  this.scrapeResult = {};
}

com.forward.invisiblehand.currentPage.prototype = {

  price: function(){
    return this.scrapeResult.price;
  },

  fixUrlForPlay: function (url) {
    if (url.match(/play\.com.+(add|delete)=\d+/)) {
      url = url.replace(/(add|delete)=\d+/, '');
    }
    if (url.match(/expansys\.com.+(sbadd)=\d+/)){
      url = url.replace(/(sbadd)=\d+/, '');
    }
    return url;
  },

  invalidGenericScrape: function(scrape){
    return !scrape || !(scrape.asin || scrape.product_query) || !scrape.region
  },

  currentScrapeFailed: function(retailer, scrapeResult) {
    if (retailer.category == "search_engine")
      return !scrapeResult || !(scrapeResult.asin || scrapeResult.product_query);
    if (retailer.category == "generic")
      return this.invalidGenericScrape(scrapeResult);
    return !scrapeResult || !(scrapeResult.price || scrapeResult.title);
  },

  scrapePage: function(scrapingCallback) {
    var self = this;
    var scrapeHandler = function(html){
      var scraper = self.extension.htmlScraper(html, self.retailer.scraper);
      scraper.scrape(function(scrapeResult) {
        scrapingCallback(scrapeResult, html);
      });
    };;

    if (this.retailer.use_inner_html) {
      this.browser.getInnerHtml(scrapeHandler);
    } else {
      this.browser.getUrl(this.url, scrapeHandler);
    }
  },

  toJson: function(){
    return JSON.stringify(this.scrapeResult);
  },

  scrape: function(callback) {
    if (!this.retailer) {
      return;
    }

    if (this.retailer.category == "search_engine" && !this.options['products-search-engines-enabled']) {
      com.forward.invisiblehand.log("Search engines have been disabled");
      return;
    }

    var self = this;
    var scrapeHandler = function(scrapeResult) {
      self.scrapeResult = scrapeResult;
      callback && callback(self.currentScrapeFailed(self.retailer, scrapeResult), self);
    };

    this.scrapePage(scrapeHandler);
  }

};
com.forward.invisiblehand.retailers = function (browser, domain) {
    this.browser = browser;
    this.retailersUrl = 'https://scrapers.' + domain + '/retailers';
    this.retailersCache = {};
}

com.forward.invisiblehand.retailers.prototype = {

  parseProductsData: function(parsedReply) {
    var regions = parsedReply.regions;
    var retailersList = this.parseRetailers(parsedReply.all_retailers, regions);
    com.forward.invisiblehand.log('Got: ' + retailersList.length + ' retailers');
    var productsData = {"regions" : regions, "genericRetailerObject" : parsedReply.generic, "retailersList" : retailersList}
    this.browser.cache("productsData", productsData);
    this.productsData = productsData;
  },

  fetchRetailers: function (callback) {
    com.forward.invisiblehand.log("Fetching retailers from server");
    var self = this;
    if (this.browser.browserType() != 'opera') { // special condition for opera
      this.browser.getUrl(this.retailersUrl, function(data) {
        self.parseProductsData(JSON.parse(data));
        if (callback) callback(self.retailers());
      });
    } else { // opera, taking scrapers from file
      this.parseProductsData(com.forward.scrapers.retailers);
      if (callback) callback(this.retailers());
    }
  },

  parseRetailers: function(fullRetailersList, regions){
    var retailersList = []
    for (var i=0; i < fullRetailersList.length; i++) {
      var retailer = fullRetailersList[i];
      if (retailer.region) {
        retailer.region = regions[retailer.region.toLowerCase()];
      }
      retailersList.push(retailer);
    }
    return retailersList;
  },

  regionFor: function(region_code){
    return this.productsData.regions[region_code];
  },

  retailers: function() {
    return this.productsData.retailersList;
  },

  genericRetailer: function() {
    return this.productsData.genericRetailerObject;
  },

  getRetailer: function(address, currentPage) {
    var region_code = currentPage ? currentPage.scrapeResult.region : undefined;
    address = address || this.browser.currentUrl();
    var retailer = this.findRetailer(address);
    if (!retailer && (address == this.browser.currentUrl())) {
      retailer = this.genericRetailer();
    }
    if (retailer && !retailer.region && region_code) {
      retailer.region = this.regionFor(region_code);
    }
    return retailer;
  },

  findRetailer: function(address){
    if (this.retailersCache[address]) return this.retailersCache[address];

    var retailers = this.retailers();
    if (!retailers) throw "The list of retailers must already have been fetched";
    for (var i = 0; i < retailers.length; i++) {
      var r = retailers[i];
      if (address.match(new RegExp(r.regex), 'i')) {
        this.retailersCache[address] = r;
        return r;
      }
    }
  },

  fetchCachedRetailers: function(callback) {
    var self = this;
    this.browser.cache("productsData", function(productsData) {
      if (productsData) {
        self.productsData = productsData;
        callback && callback(self.retailers());
      } else {
        self.fetchRetailers(callback);
      }
    });
  }

};
(function() {
  var ProductsIframe,
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; };

  ProductsIframe = (function(_super) {

    __extends(ProductsIframe, _super);

    ProductsIframe.name = 'ProductsIframe';

    function ProductsIframe() {
      return ProductsIframe.__super__.constructor.apply(this, arguments);
    }

    ProductsIframe.prototype.initialize = function(retailers) {
      var message, reporting,
        _this = this;
      this.retailers = retailers;
      this.insert();
      reporting = {
        src: this.extension.installationSource(),
        retailer_name: this.currentPage.retailer.name,
        category: this.currentPage.retailer.category,
        uid: this.extension.properties.uid,
        browser: this.extension.browser.browserType(),
        extension: this.extension.browser.extensionVersion()
      };
      message = {
        url: this.currentPage.url,
        scrape: this.currentPage.scrapeResult,
        region: this.currentPage.retailer.region,
        reporting: reporting,
        settingsLink: this.extension.properties.settingsLink
      };
      this.messaging.setMessageHandler("scrape", function(msg) {
        return _this.scrape(msg.url);
      });
      return this.notify("initialize", message);
    };

    ProductsIframe.prototype.scrape = function(url) {
      var retailer,
        _this = this;
      retailer = this.retailers.getRetailer(url);
      if (retailer) {
        return new com.forward.invisiblehand.RemoteScrape(this.extension).scrape(url, retailer, function(scrape) {
          return _this.notify("scrapeResult", {
            url: url,
            scrape: scrape
          });
        });
      }
    };

    return ProductsIframe;

  })(com.forward.invisiblehand.Iframe);

  com.forward.invisiblehand.ProductsIframe = ProductsIframe;

}).call(this);
com.forward.invisiblehand.products = function (extension, options) {
  var domain = extension.properties.domain || "invisiblehand.co.uk";
  this.options = options;
  this.extension = extension;
  this.browser = extension.browser;
  this.retailers = new com.forward.invisiblehand.retailers(extension.browser, domain);
}

com.forward.invisiblehand.products.prototype = {

  scrapeCurrentPage: function(callback) {
    var currentPage = new com.forward.invisiblehand.currentPage(this.retailers, this.options, this.browser, this.extension, this.relatedProducts);
    currentPage.scrape(callback);
  },

  run: function(mixpanelTrackingCallback) {
    var self = this;
    self.retailers.fetchCachedRetailers(function(retailers) {
      self.scrapeCurrentPage(function(err, currentPage) {
        if (err) {
          com.forward.invisiblehand.log("Scraping current page failed");
        } else {
          var domain = self.extension.properties.domain || "invisiblehand.co.uk";
          var iFrameAddress = "http://productsiframe2." + domain + "/?tracking_code=" + self.extension.properties.trackingCode;
          var productsIframe = new com.forward.invisiblehand.ProductsIframe("invisiblehand-iframe", iFrameAddress, self.extension, currentPage);
          productsIframe.initialize(self.retailers);
        }
      });
    });
  }
};

com.forward.invisiblehand.travel_operators = {}

com.forward.invisiblehand.flights = function (extension, options) {
  var domain = extension.properties.domain || "invisiblehand.co.uk";
  this.extension = extension;
  this.browser = extension.browser;
  this.travelOperatorsUrl = 'https://scrapers.invisiblehand.co.uk/travel_operators';
  this.iFrameAddress = 'fly.' + domain;
  this.options = options;
}

com.forward.invisiblehand.flights.prototype = {

  travelOperators: function() {
    return this.travelOperatorsList;
  },

  parseTravelOperatorsData: function(parsedReply) {
    this.travelOperatorsList = parsedReply.operators;
    for (var i=0; i < this.travelOperatorsList.length; i++) {
      if (this.travelOperatorsList.region) this.travelOperatorsList.region = parsedReply.regions[this.travelOperatorsList.region.toLowerCase()];
    }
    com.forward.invisiblehand.log('Got: ' + this.travelOperatorsList.length + ' travel operators');
    this.browser.cache("travelOperatorsList", this.travelOperatorsList);
  },

  fetchTravelOperators: function(callback) {
    if (com.forward.invisiblehand.travel_operators.scrapers) {
      this.travelOperatorsList = com.forward.invisiblehand.travel_operators.scrapers.operators;
      callback(this.travelOperatorsList);
      return;
    }

    var self = this;
    var travelOperatorsList = this.travelOperators();
    if (!travelOperatorsList) {
      if (this.browser.browserType() != 'opera') { // special condition for opera
        this.browser.getUrl(this.travelOperatorsUrl, function(data) {
          self.parseTravelOperatorsData(JSON.parse(data));
          if (callback) callback(self.travelOperatorsList);
        });
      } else { // opera, taking scrapers from file
        this.parseTravelOperatorsData(com.forward.scrapers.travel_operators);
        if (callback) callback(this.travelOperatorsList);
      }
    } else {
      if (callback) callback(travelOperatorsList);
    }
  },

  consolidateStops: function(scrape) {
    if (!scrape) return;
    if (scrape.outbound_stops != null || scrape.return_stops != null)
      scrape.stops = (parseInt(scrape.outbound_stops) || 0) + (parseInt(scrape.return_stops) || 0);
  },

  scrapeCurrentPage: function(callback) {
    var self = this;
    var travelOperator = this.getTravelOperator();
    if (!travelOperator) {
      com.forward.invisiblehand.log("This is not a travel operator, skipping");
      return;
    }
    com.forward.invisiblehand.log("Scraping the travel operator");
    var scraper = this.extension.domScraper(travelOperator.scraper);
    scraper.scrape(function (scrape) {
      self.consolidateStops(scrape);
      com.forward.invisiblehand.log("Scraped: ");
      com.forward.invisiblehand.log(JSON.stringify(scrape));
      callback(self.processScrapeResults(scrape));
    });
  },

  processScrapeResults: function(scrape) {
    if (!scrape) return;
    if (!scrape.currency) {
      scrape.currency = this.detectCurrency(scrape.price);
    }
    scrape.price = this.cleanPrice(scrape.price);
    return scrape;
  },

  cleanPrice: function(price) {
    if (typeof price != 'string') return price;

    return price.replace(/[^\d\.]/g, "");
  },

  detectCurrency: function (price) {
    if (typeof price != 'string') return;

    var currencyMatches = {
      USD: [/USD/, /\$/],
      EUR: [/EUR/, /€/],
      GBP: [/GBP/, /£/]
    };

    for (currency in currencyMatches)
      for (var i = 0; i < currencyMatches[currency].length; i++)
        if (price.match(currencyMatches[currency][i]))
          return currency;


    var unknownCurrencyMatch = price.match(/[\d\.,]\s*([A-Z]{3})/) || price.match(/([A-Z]{3})\s*[,\d\.]+/);
    if (unknownCurrencyMatch) return unknownCurrencyMatch[1];

    return null;
  },

  getTravelOperator: function(address) {
    address = address || this.browser.currentUrl();
    var travelOperators = this.travelOperators();
    if (!travelOperators) throw "The list of travel operators must already have been fetched";
    for (var i = 0; i < travelOperators.length; i++) {
      var travelOperator = travelOperators[i];
      if (address.match(new RegExp(travelOperator.regex), 'i')) {
        return travelOperator;
      }
    }
    return null;
  },

  showIFrame: function(scrape, address) {
    var travelOperator = this.getTravelOperator();

    var params = scrape;
    params.page_url       = this.browser.currentUrl();
    params.browser        = this.browser.browserType();
    params.window_height  = this.browser.windowHeight();
    params.src            = this.extension.properties.source;
    params.tracking_code  = this.extension.properties.trackingCode;
    params.acf              = this.extension.properties.uid; // acf is random and is just so that 'uid' doesn't appear on the URL

    var iframeUrl = com.forward.invisiblehand.common.utils.buildUrl(this.iFrameAddress + '/' + address, params);

    var currentPage = {
      retailer: travelOperator,
      scrapeResult: scrape,
      url: this.browser.currentUrl()
    }
    var iframe = new com.forward.invisiblehand.Iframe("invisiblehand-iframe", iframeUrl, this.extension, currentPage);
    iframe.insert();
  },

  canShowNotification: function(scrape) {
    return scrape.depart_from && scrape.destination && scrape.departure_date; // scrape is undefined at times, do something about it
  },

  canShowSearchForm: function(scrape) {
    return scrape.depart_from || scrape.destination;
  },

  run: function() {
    var self = this;
    if (this.options['flights-enabled'] === false) {
      com.forward.invisiblehand.log("Flights are disabled");
      return;
    }
	this.browser.cache('travelOperatorsList', function(travelOperatorsList) {
	    self.travelOperatorsList = travelOperatorsList;
	   	self.fetchTravelOperators(function() {
	      self.scrapeCurrentPage(function(scrape) { // clean the results?
	        if (scrape) {
	          if (self.canShowNotification(scrape)) return self.showIFrame(scrape, 'notification');
	          if (self.canShowSearchForm(scrape))   return self.showIFrame(scrape, 'search');
	        }
	        com.forward.invisiblehand.log("Scraping of flights failed");
	      })
	    })
	});
    return true;
  }
}
com.forward.invisiblehand.hotels = function (extension, options) {
  var domain = extension.properties.domain || "invisiblehand.co.uk";
  this.extension = extension;
  this.browser = extension.browser;
  this.hotelsUrl = 'https://scrapers.invisiblehand.co.uk/hotels';
  this.iFrameAddress = 'hotels-iframe.' + domain;
  this.options = options;
}

com.forward.invisiblehand.hotels.prototype = {

  travelOperators: function() {
    return this.hotelsList;
  },

  parseHotelsData: function(parsedReply) {
    this.hotelsList = parsedReply.operators;
    for (var i=0; i < this.hotelsList.length; i++) {
      if (this.hotelsList.region) this.hotelsList.region = parsedReply.regions[this.hotelsList.region.toLowerCase()];
    }
    com.forward.invisiblehand.log('Got: ' + this.hotelsList.length + ' hotel sites');
    this.browser.cache("hotelsList", this.hotelsList);
  },

  fetchTravelOperators: function(callback) {
    if (com.forward.invisiblehand.hotels.scrapers) {
      this.hotelsList = com.forward.invisiblehand.hotels.scrapers.operators;
      callback(this.hotelsList);
      return;
    }

    var self = this;
    var hotelsList = this.travelOperators();
    if (!hotelsList) {
      if (this.browser.browserType() != 'opera') { // special condition for opera
        this.browser.getUrl(this.hotelsUrl, function(data) {
          self.parseHotelsData(JSON.parse(data));
          if (callback) callback(self.hotelsList);
        });
      } else { // opera, taking scrapers from file
        this.parseHotelsData(com.forward.scrapers.hotels);
        if (callback) callback(this.hotelsList);
      }
    } else {
      if (callback) callback(hotelsList);
    }
  },

  scrapeCurrentPage: function(callback) {
    var self = this;
    var travelOperator = this.getTravelOperator();
    if (!travelOperator) {
      com.forward.invisiblehand.log("This is not a hotel site, skipping");
      return;
    }
    com.forward.invisiblehand.log("Scraping the hotel site");
    var scraper = this.extension.domScraper(travelOperator.scraper);
    scraper.scrape(function (scrape) {
      com.forward.invisiblehand.log("Scraped: ");
      com.forward.invisiblehand.log(scrape);
      callback(self.processScrapeResults(scrape));
    });
  },

  processScrapeResults: function(scrape) {
    if (!scrape) return;
    return scrape;
  },

  getTravelOperator: function(address) {
    address = address || this.browser.currentUrl();
    var travelOperators = this.travelOperators();
    if (!travelOperators) throw "The list of travel operators must already have been fetched";
    for (var i = 0; i < travelOperators.length; i++) {
      var travelOperator = travelOperators[i];
      if (address.match(new RegExp(travelOperator.regex), 'i')) {
        return travelOperator;
      }
    }
    return null;
  },

  overridingStylesheet: function () {
    var travelOperator = this.getTravelOperator();
    return travelOperator.style_override;
  },

  showIFrame: function(scrape, address) {
    var travelOperator = this.getTravelOperator();

    var params = scrape;
    params.page_url       = this.browser.currentUrl();
    params.browser        = this.browser.browserType();
    params.region         = scrape.region || travelOperator.region;
    params.window_height  = this.browser.windowHeight();
    params.src            = this.extension.properties.source;
    params.tracking_code  = this.extension.properties.trackingCode;
    params.acf            = this.extension.properties.uid;
    params.site           = travelOperator.name;

    var iframeUrl = com.forward.invisiblehand.common.utils.buildUrl(this.iFrameAddress + '/' + address, params);
    var message = { uid: params.acf, browser: params.browser, src: params.src, tracking_code: params.tracking_code, site: travelOperator.name, region: params.region };

    var currentPage = {
      retailer: travelOperator,
      scrapeResult: scrape,
      url: this.browser.currentUrl()
    }
    var iframe = new com.forward.invisiblehand.Iframe("invisiblehand-iframe", iframeUrl, this.extension, currentPage);
    iframe.insert();
    iframe.notify("initialize", message);
  },

  canShowNotification: function(scrape) {
    return scrape.checkin_date && scrape.checkout_date && scrape.location;
  },

  run: function() {
    var self = this;
    if (!this.options['flights-enabled']) {
      com.forward.invisiblehand.log("Flights (and hotels) are disabled");
      return;
    }
    this.browser.cache('hotelsList', function(hotelsList) {
      self.hotelsList = hotelsList;
      self.fetchTravelOperators(function() {
        self.scrapeCurrentPage(function(scrape) { // clean the results?
          if (scrape) {
            if (self.canShowNotification(scrape)) return self.showIFrame(scrape, 'notification');
          }
          com.forward.invisiblehand.log("Scraping of hotels failed");
        })
      })
    });
    return true;
  }
}
com.forward.invisiblehand.rentals = function (extension, options) {
  var domain = extension.properties.domain || "invisiblehand.co.uk";
  this.extension = extension;
  this.browser = extension.browser;
  this.rentalsUrl = 'https://scrapers.invisiblehand.co.uk/rentals';
  this.iFrameAddress = 'rentals-iframe.' + domain;
  this.options = options;
}

com.forward.invisiblehand.rentals.prototype = {

  travelOperators: function() {
    return this.rentalsList;
  },

  parseRentalsData: function(parsedReply) {
    this.rentalsList = parsedReply.operators;
    for (var i=0; i < this.rentalsList.length; i++) {
      if (this.rentalsList.region) this.rentalsList.region = parsedReply.regions[this.rentalsList.region.toLowerCase()];
    }
    com.forward.invisiblehand.log('Got: ' + this.rentalsList.length + ' rental sites');
    this.browser.cache("rentalsList", this.rentalsList);
  },

  fetchTravelOperators: function(callback) {
    if (com.forward.invisiblehand.rentals.scrapers) {
      this.rentalsList = com.forward.invisiblehand.rentals.scrapers.operators;
      callback(this.rentalsList);
      return;
    }

    var self = this;
    var rentalsList = this.travelOperators();
    if (!rentalsList) {
      if (this.browser.browserType() != 'opera') { // special condition for opera
        this.browser.getUrl(this.rentalsUrl, function(data) {
          self.parseRentalsData(JSON.parse(data));
          if (callback) callback(self.rentalsList);
        });
      } else { // opera, taking scrapers from file
        this.parseRentalsData(com.forward.scrapers.rentals);
        if (callback) callback(this.rentalsList);
      }
    } else {
      if (callback) callback(rentalsList);
    }
  },

  scrapeCurrentPage: function(callback) {
    var self = this;
    var travelOperator = this.getTravelOperator();
    if (!travelOperator) {
      com.forward.invisiblehand.log("This is not a rental site, skipping");
      return;
    }
    com.forward.invisiblehand.log("Scraping the rental site");
    var scraper = this.extension.domScraper(travelOperator.scraper);
    scraper.scrape(function (scrape) {
      com.forward.invisiblehand.log("Scraped: ");
      com.forward.invisiblehand.log(scrape);
      callback(scrape);
    });
  },

  getTravelOperator: function(address) {
    address = address || this.browser.currentUrl();
    var travelOperators = this.travelOperators();
    if (!travelOperators) throw "The list of travel operators must already have been fetched";
    for (var i = 0; i < travelOperators.length; i++) {
      var travelOperator = travelOperators[i];
      if (address.match(new RegExp(travelOperator.regex), 'i')) {
        return travelOperator;
      }
    }
    return null;
  },

  overridingStylesheet: function () {
    var travelOperator = this.getTravelOperator();
    return travelOperator.style_override;
  },

  showIFrame: function(scrape) {
    var travelOperator = this.getTravelOperator();

    var params = scrape;
    params.region         = travelOperator.region;
    params.page_url       = this.browser.currentUrl();
    params.browser        = this.browser.browserType();
    params.window_height  = this.browser.windowHeight();
    params.src            = this.extension.properties.source;
    params.tracking_code  = this.extension.properties.trackingCode;
    params.acf            = this.extension.properties.uid;

    var iframeUrl = com.forward.invisiblehand.common.utils.buildUrl(this.iFrameAddress + '/notification', params);
    var message = { uid: params.acf, browser: params.browser, src: params.src, tracking_code: params.tracking_code, site: travelOperator.name };

    var currentPage = {
      retailer: travelOperator,
      scrapeResult: scrape,
      url: this.browser.currentUrl()
    }
    var iframe = new com.forward.invisiblehand.Iframe("invisiblehand-iframe", iframeUrl, this.extension, currentPage);
    iframe.insert();
    iframe.notify("initialize", message);
  },

  canShowNotification: function(scrape) {
    return scrape.checkin_date;
  },

  run: function() {
    var self = this;
    if (this.options['rentals-enabled'] === false) {
      com.forward.invisiblehand.log("Rentals are disabled");
      return;
    }
    this.browser.cache('rentalsList', function(rentalsList) {
      self.rentalsList = rentalsList;
      self.fetchTravelOperators(function() {
        self.scrapeCurrentPage(function(scrape) { // clean the results?
          if (scrape) {
            if (self.canShowNotification(scrape)) return self.showIFrame(scrape);
          }
          com.forward.invisiblehand.log("Scraping of rentals failed");
        })
      })
    });
    return true;
  }
}
com.forward.invisiblehand.tickets = function (extension, options) {
  var domain = extension.properties.domain || "invisiblehand.co.uk";
  this.extension = extension;
  this.browser = extension.browser;
  this.ticketsUrl = 'https://scrapers.invisiblehand.co.uk/tickets';
  this.iFrameAddress = 'tickets-iframe.' + domain;
  this.options = options;
}

com.forward.invisiblehand.tickets.prototype = {

  ticketOperators: function() {
    return this.ticketsList;
  },

  parseTicketsData: function(parsedReply) {
    this.ticketsList = parsedReply.operators;
    for (var i=0; i < this.ticketsList.length; i++) {
      if (this.ticketsList.region) this.ticketsList.region = parsedReply.regions[this.ticketsList.region.toLowerCase()];
    }
    com.forward.invisiblehand.log('Got: ' + this.ticketsList.length + ' ticket sites');
    this.browser.cache("ticketsList", this.ticketsList);
  },

  fetchTicketOperators: function(callback) {
    if (com.forward.invisiblehand.tickets.scrapers) {
      this.ticketsList = com.forward.invisiblehand.tickets.scrapers.operators;
      callback(this.ticketsList);
      return;
    }

    var self = this;
    var ticketsList = this.ticketOperators();
    if (!ticketsList) {
      if (this.browser.browserType() != 'opera') { // special condition for opera
        this.browser.getUrl(this.ticketsUrl, function(data) {
          self.parseTicketsData(JSON.parse(data));
          if (callback) callback(self.ticketsList);
        });
      } else { // opera, taking scrapers from file
        this.parseTicketsData(com.forward.scrapers.tickets);
        if (callback) callback(this.ticketsList);
      }
    } else {
      if (callback) callback(ticketsList);
    }
  },

  scrapeCurrentPage: function(callback) {
    var self = this;
    var ticketOperator = this.getTravelOperator();
    if (!ticketOperator) {
      com.forward.invisiblehand.log("This is not a ticket site, skipping");
      return;
    }
    com.forward.invisiblehand.log("Scraping the ticket site");
    var scraper = this.extension.domScraper(ticketOperator.scraper);
    scraper.scrape(function (scrape) {
      com.forward.invisiblehand.log("Scraped: ");
      com.forward.invisiblehand.log(scrape);
      callback(scrape);
    });
  },

  getTravelOperator: function(address) {
    address = address || this.browser.currentUrl();
    var ticketOperators = this.ticketOperators();
    if (!ticketOperators) throw "The list of ticket operators must already have been fetched";
    for (var i = 0; i < ticketOperators.length; i++) {
      var ticketOperator = ticketOperators[i];
      if (address.match(new RegExp(ticketOperator.regex), 'i')) {
        return ticketOperator;
      }
    }
    return null;
  },

  overridingStylesheet: function () {
    var ticketOperator = this.getTravelOperator();
    return ticketOperator.style_override;
  },

  showIFrame: function(scrape) {
    var ticketOperator = this.getTravelOperator();

    var params = scrape;
    params.region         = ticketOperator.region;
    params.page_url       = this.browser.currentUrl();
    params.browser        = this.browser.browserType();
    params.window_height  = this.browser.windowHeight();
    params.src            = this.extension.properties.source;
    params.tracking_code  = this.extension.properties.trackingCode;
    params.acf            = this.extension.properties.uid;

    var iframeUrl = com.forward.invisiblehand.common.utils.buildUrl(this.iFrameAddress + '/notification', params);
    var message = { uid: params.acf, browser: params.browser, src: params.src, tracking_code: params.tracking_code, site: ticketOperator.name };

    var currentPage = {
      retailer: ticketOperator,
      scrapeResult: scrape,
      url: this.browser.currentUrl()
    }
    var iframe = new com.forward.invisiblehand.Iframe("invisiblehand-iframe", iframeUrl, this.extension, currentPage);
    iframe.insert();
    iframe.notify("initialize", message);
  },

  canShowNotification: function(scrape) {
    return scrape.date && scrape.artist;
  },

  run: function() {
    var self = this;
    if (this.options['tickets-enabled'] === false) {
      com.forward.invisiblehand.log("tickets are disabled");
      return;
    }
    this.browser.cache('ticketsList', function(ticketsList) {
      self.ticketsList = ticketsList;
      self.fetchTicketOperators(function() {
        self.scrapeCurrentPage(function(scrape) { // clean the results?
          if (scrape) {
            if (self.canShowNotification(scrape)) return self.showIFrame(scrape);
          }
          com.forward.invisiblehand.log("Scraping of tickets failed");
        })
      })
    });
    return true;
  }
}
com.forward.invisiblehand.domainWhitelist = ["tools-plus.com","911healthshop.com","target.com","simplyelectronics.net","schlecker.com","mytoys.de","schwab.de","shoes.com","wwbw.com","plumbnation.co.uk","tower.com","ajelectronics.co.uk","yahoo.com","mothernature.com","4electronicwarehouse.com","6ave.com","marksandspencer-appliances.com","ebags.com","porzellantreff.de","soap.com","baur.de","jessops.com","cafepress.com","petmountain.com","picstop.co.uk","fragrancenet.com","homedepot.com","ebay.com","game.co.uk","argos.co.uk","1stoplighting.com","GlassesUsa.com","amazon.ca","petvillas.co.uk","elitecarseats.com","snowshack.com","jeanswelt.de","kiddisave.co.uk","dvdplanet.com","toysrus.com","alphamusic.de","braceshop.com","radioshack.com","crutchfield.com","1800pools.com","buecher.de","sonystyle.com","ukbuy.com","onplumb.com","sonystyle.ca","misco.co.uk","ebay.de","restockit.com","otto.de","statelinetack.com","bestbuyeyeglasses.com","hagebau.de","destinationlighting.com","pcconnectionexpress.com","healthchecksystems.com","cabuy.com","moosejaw.com","efaucets.com","kodak.com","pcworld.co.uk","mrwatch.com","scrapbookingalley.com","directcosmetics.com","petsupplies.com","appliancedeals.co.uk","etronics.com","sears.com","wirefly.com","discountgolfworld.com","ccmusic.com","lightingcatalog.com","homedepot.ca","theperfumespot.com","shopbt.com","dog.com","homelement.com","fashionbuglanebryant.com","shoplet.com","douglas.de","zipzoomfly.com","horse.com","tooled-up.com","luggagepoint.com","chaptersindigo.ca","comfortfirst.com","footsmart.com","bookshopblackwell.co.uk","mymemory.co.uk","shopadvanceautoparts.com","electronics-expo.com","pcwb.co.uk","printronic.com","redcoon.de","hitseller.de","bol.de","golfgalaxy.com","irvsluggage.com","compuvest.com","fragrancex.com","masseysoutfitters.com","joylot.com","superwarehouse.com","modernfurniturewarehouse.com","uksearchyahoo.com","very.co.uk","dixons.co.uk","currys.co.uk","onsale.com","pharmacy2u.co.uk","pcmall.com","watchwear.com","macwarehouse.co.uk","asda-pharmacy.co.uk","kolibrishop.com","thatpetplace.com","hollywoodtoysandcostumes.com","shindigz.com","jimmyjazz.com","leekes.co.uk","shopgetorganized.com","perfumeemporium.com","air-n-water.com","odoroso.com","computers4sure.com","rebuy.de","hitflip.de","cyberport.de","medpex.de","tigerdirect.ca","cotendo.com","entirelypets.com","scentiments.com","discountwatchstore.com","compuplus.com","ftd.com","orangeonions.com","carolinarustica.com","studica.com","interstatemusic.com","diamondharmony.com","nextdomesticappliances.co.uk","ecost.com","thenerds.net","andysautosport.com","coveroo.com","biggerbooks.com","modernfurniture4home.com","zeilenreich.de","brookstone.com","johnlewis.com","watchshopuk.com","bellacor.com","the-house.com","guitarcenter.com","hmv.com","diytools.co.uk","chemistdirect.co.uk","bing.com","meijer.com","ccvideo.com","pets-warehouse.com","phase-eight.co.uk","glassesusa.com","cymax.com","dealyard.com","vanns.com","blue-tomato.com","athisbest.com","shopperschoice.com","musicfactorydirect.com","laptopsdirect.co.uk","kitchensource.com","costumes4less.com","pixmania.de","a1books.com","sexualwellbeing.com","vitaminmenu.com","redstamp.com","adzanox.com","betterworldbooks.com","contentvcommerce.com","searsoutlet-new.com","buch.de","bergners.com","monoprice.com","seapets.co.uk","missionrs.com","morphyrichards.co.uk","sebago.com","cotswoldoutdoor.com","fun.co.uk","newegg.ca","techdepot.com","apple.com/uk","apple.com/us","cymaxstores.com","sole.de","ebay.ca","247electrical.co.uk","gymcompany.co.uk","mobilefun.co.uk","kiddies-kingdom.com","inkfactory.com","microdirect.co.uk","quizclothing.co.uk","prepaymania.co.uk","choicesuk.com","electrical123.com","goldsmiths.co.uk","pcworldbusiness.co.uk","sportolino.de","largejeans.de","dnet24.de","unbeatablesaleaffiliatetechnology.com","mobilejoker.de","gogoritas.com","quickaudio.de","koffer-direkt.de","trade-a-game.de","elektromall.de","elektropoint24.de","netzwerkartikel.de","eshopbrokers.co.uk","voelkner.de","shoppingpoint24.com","porzellanhandel24.de","hoodboyz.de","wwwhoodboyz.de","spielgeschenke.de","trendor.de","play.com","puzzle-offensive.de","worldofproduct.de","handwerker-versand.de","oyyy.co.uk","focalprice.com","easycosmetic.de","streetsideauto.com","zavvi.com","onestopfanshop.com","cruisercustomizing.com","chickdowntown.com","productexpress.com","cherrybrook.com","printus24.de","bagking.com","containerstore.com","beck-shop.de","petshopusa.com","vividracing.com","mercantila.com","cafepress.co.uk","shopvisiondecor.com","testequipmentdepot.com","babyearth.com","livingroomwarehouse.com","academicsuperstore.com","elightbulbs.com","shopchristmascentral.com","crystal-co.com","fansedge.com","medexsupply.com","biggerbras.com","mystrollers.com","greekgear.com","eshopmexx.com","toysrus.ca","lightingbygregory.com","etoys.com","bagsbuy.com","sportsmemorabilia.com","bargainstation.com","tvcables.co.uk","sweatband.com","hbo.com","autopartswarehouse.com","outletaltrec.com","stylintrucks.com","streetbeatcustoms.com","sears.ca","thewalkingcompany.com","gemvara.com","mydesignshop.com","samsclub.com","4wheelparts.com","gandermountain.com","tennisexpress.com","mwave.com","webstaurantstore.com","garden.com","rockymountaintrail.com","jomashop.com","isave.com","raffaello-network.com","rockymountainatvmc.com","rugsale.com","sainsburys.co.uk","officesupersavers.com","thegreatindoors.com","taylorgifts.com","zoobuy.net","shophistory.com","cengagebrain.co.uk","gbdollarnights.com","catherineslanebryant.com","qvc.com","planetshoes.com","usallsaints.com","mcmelectronics.com","stumpsparty.com","teachercreated.com","tiptoeturtle.com","totalvac.com","hifi-tower.co.uk","tom-tailor.de","reifen.com","yoox.com","macys.com","bravado.de","conrad.de","quickshopping.de","tophandy.de","d-living.de","doit24.de","parfumdreams.de","getmobile.de","fahrrad.de","wolfgangsvault.com","footlocker.com","autoaccessoriesgarage.com","thesource.ca","myjewelrybox.com","atdouglas-shop.com","hagelshop.de","shopfussball.de","lebensmittel.de","imwalking.de","nbcuniversalstore.com","petsmart.com","zooroyal.de","buero-direkt.de","taschenkaufhaus.de","3kisseslingerie.com","connox.de","manalive.com","shopbeate-uhse.com","onlynaturalpet.com","starcostumes.com","inkgrabber.com","rockler.com","seaofdiamonds.com","americandiabeteswholesale.com","getabstract.com","metrokitchen.com","aspenpublishers.com","llewellyn.com","pasttimes.com","athletagap.com","golfetail.com","sitstay.com","shopecko.com","worldofwatches.com","tnvitamins.com","prosportsmemorabilia.com","oriononlineshop.com","chinavasion.com","petstore.com","plus.de","pacificgeek.com","t-online-shop.de","splashredcatsusa.com","homeandpatiodecorcenter.com","gandoubleclick.net","zooplus.de","comfortmarket.com","telescope.com","skinit.com","adultshop.de","forzieri.com","homelife24.com","elektronik-star.de","babyviva.com","collectablesdirect.com","collectablesdirect.ca","time2.co.uk","rockbottomgolf.com","onestepahead.com","redenvelope.com","thedrumworks.com","shoebuy.com","bedroomfurnituremart.com","jamesallen.com","allvitaminsplus.com","builddirect.com","zephyrsports.com","wholesalecostumeclub.com","luggageguru.com","patioumbrellas.com","healthykin.com","overtons.com","perfume-worldwide.com","folica.com","willygoat.com","ereo.com","iwantoneofthose.com","thekitchenappliancestore.co.uk","conrad-uk.com","bhsdirect.co.uk","ukinsight.com","businessitonline.com","officestationery.co.uk","crescentelectronics.co.uk","northerntooluk.com","kitchenscience.co.uk","bambinodirect.co.uk","smartbuyglasses.co.uk","laskys.com","mainzoo.de","spieletraeume.de","spiele-offensive.de","asda-entertainment.co.uk","hq-patronen.de","kaiserstuhlshop.de","cpotools.com","heartlandamerica.com","jr.com","funtoymall.com","toolschest.com","barnesandnoble.com","gettington.com","vitaminemporium.com","perfume.com","allergybegone.com","gameseek.co.uk","birthdayinabox.com","thingsremembered.com","buyz.com","javari.co.uk","endless.com","ebuyer.com","bhphotovideo.com","amazon.de","capitolsupply.com","bookdepository.co.uk","expansys.com","amazon.co.uk","pixmania.co.uk","abt.com","equestriancollections.com","ebay.co.uk","dabs.com","onestopplus.com","lifeuniform.com","bookbyte.com","google.de","scan.co.uk","factoryoutletstore.com","sunnysports.com","base.com","plantmenow.co.uk","philipmorrisdirect.co.uk","schuh.co.uk","bestbuy.com","ashford.com","buy.com","borders.com","amazon.com","solidsignal.com","bestbuy.co.uk","lowes.com","google.com","google.co.uk","ryman.co.uk","halfords.com","number1direct.com","diapers.com","singer22.com","karstadt.de","sundiego.com","roadrunnersports.com","cclonline.com","butterflyphoto.com","circuitcity.com","compusa.com","instawares.com","allergysuperstore.com","walmart.com","unbeatablesale.com","beallsflorida.com","electricshopping.com","wateroutfitters.com","watchmad.co.uk","warehouseexpress.com","mywebsearch.com","cosmeticamerica.com","cosmeticmall.com","outdoorfurnituregallery.com","music123.com","sobongo.com","macmall.com","tescoentertainment.com","direct.tesco.com","thehut.com","moviemars.com","viking-direct.co.uk","altrec.com","appliancesonline.co.uk","1800lighting.com","chefgiant.com","boots.com","kenmarwatches.com","lambda-tek.com","housershoes.com","leftlanesports.com","nordstrom.com","ritzcamera.com","overstock.com","e2save.com","boatersworld.com","hughesdirect.co.uk","executiveessentials.com","teilesuche24.de","bulbamerica.com","cardstore.com","gaiam.com","discovery.com","drugstore.com","dazadi.com","gamefly.com","wowhd.co.uk","Swiminn.com","pcrush.com","heatandplumb.com","abebooks.de","fireforless.com","pugster.com","officemax.com","costumecraze.com","kmart.com","appliancesdirect.co.uk","feelunique.com","skymall.com","brylanehome.com","newegg.com","zalando.co.uk","visiondirect.com","cuisinartwebstore.com","mediaversand.de","musiciansfriend.com","ravensburger.de","everythingplay.com","hughedirect.co.uk","competitiveheatingsupplies.co.uk","officegiant.co.uk","autopartsexpress.com","homeclick.com","yourgolftravel.com","herroom.com","photo4less.com","bostonstore.com","carsons.com","younkers.com","secondspin.com","stuff4crossstitch.com","lightingshowroom.com","etundra.com","theblackbow.com","mintprice.com","dialaphone.co.uk","thetwistergroup.com","leisurepro.com","screwfix.com","petcarerx.com","fiftyplus.co.uk","tylertool.com","motivators.com","marinedepot.com","arkivmusic.com","fathead.com","123inkjets.com","7dayshop.com","4wd.com","ubid.com","mankind.co.uk","compsource.com","comet.co.uk","officedepot.com","home-improvement-superstore.com","videocollection.com","valuepetsupplies.com","herbspro.com","escentual.com","jcwhitney.com","bicwarehouse.com","zoostores.com","lnt.com","diy.com","barratts.co.uk","drjays.com","directasda.com","geeks.com","intheholegolf.com","kimmyshop.com","ebooks.com","cameraworld.com","danskin.com","alljigsawpuzzles.co.uk","zazzle.com","wolfcamera.com","tigerdirect.com","101phones.com","microwarehouse.co.uk","abesofmaine.com","textbooks.com","sendit.com","misco.de","bbcshop.com","adorama.com","beautybar.com","halloweenmart.com","activeforever.com","eforcity.com","toysrus.co.uk","nothingbutsoftware.com","everythingfurniture.com","walgreens.com","passiontec.de","digitalo.de","fabial.de","uhrcenter.de","chefscatalog.com","ukforzieri.com","oneclickpharmacy.co.uk","yagma.de","brickhousesecurity.com","caseking.de","pgestore.com","purelygadgets.co.uk","991.com","activebeauty.co.uk","bootskitchenappliances.com","argosentertainment.co.uk","thenaturalstore.com","shoplovefilm.com","choice.co.uk","1928.com","wbshop.com","universitybooksonline.com","michaelstars.com","giftcollector.com","cduniverse.com","herbergers.com","dillards.com","cyberguys.com","deepdiscount.com","beauty.com","zumaoffice.com","laptopz.com","beautybridge.com","smartbargains.com","whatabeautifullife.com","theclassyhome.com","laptopbatteryexpress.com","mattressesworld.co.uk","uktoolcentre.co.uk","soccergarage.com","buydig.com","organize.com","shopcallawaygolf.com","gamefly.co.uk","packyourbagstravelstore.com","catfootwear.com","naturalizer.com","debenhamsplus.com","techforless.com","lookfantastic.com","crucial.com","ambrosewilson.com","giggle.com","longtallsally.com","sarenza.co.uk","djpremium.com","partypieces.co.uk","simplyhe.co.uk","lookmantastic.com","bestbullysticks.com","everythingofficefurniture.com","finishline.com","beachcamera.com","comtech.de","flirtcatalog.com","ajmadison.com","merlincycles.com","proaudiostar.com","boohoo.com","thewatchhut.co.uk","moviemail-online.co.uk","musicarts.com","ukofficedirect.co.uk","3balls.com","fish.com","viosoftware.com","urbansurfer.co.uk","swell.com","officeking.com","electricaldiscountuk.co.uk","carparts.com","amazingsocks.com","cookieskids.com","amatteroffax.com","milanoo.com","electronicexpress.com","expansys-usa.com","pcconnection.com","bonton.com","macconnection.com","thebookpeople.co.uk","novatech.co.uk","elder-beerman.com","eyesave.com","pet-supermarket.co.uk","famousfootwear.com","houseofbrides.com","yoursclothing.co.uk","uniformcity.com","ulta.com","toolking.com","tmart.com","surffanatics.com","sportsmith.net","gap.com","beautyexpert.co.uk","ecampus.com","katom.com","bhs.co.uk","modainpelle.com","footballfanatics.com","office.co.uk","dvdgold.co.uk","euroffice.co.uk","kazootoys.com","invisiblehand.co.uk"];
com.forward.invisiblehand.travel_operators.scrapers = {"regions":{"uk":{"name":"United Kingdom","code":"UK","server":"uk.lb.invisiblehand.co.uk","currency":{"name":"GBP","symbol":"\u00a3"}},"de":{"name":"Germany","code":"DE","server":"de.lb.invisiblehand.co.uk","currency":{"name":"EUR","symbol":"\u20ac"}},"ca":{"name":"Canada","code":"CA","server":"ca.lb.invisiblehand.co.uk","currency":{"name":"CAD","symbol":"$"}},"us":{"name":"United States","code":"US","server":"us.lb.invisiblehand.co.uk","currency":{"name":"USD","symbol":"$"}}},"operators":[{"region":"US","regex":"^https?://(www\\.)?virgin-atlantic\\.com/en/us.+?checkProcessingStatus","name":"VirginAtlanticComUs","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
if ($ih('.total').length > 0) {
  callback($ih('.total:first').text().replace("Total fare", ""));
  return;
} else if ($ih('.priceCol strong').length > 0) {
  callback($ih('.priceCol strong:first').text().replace(/per adult.*/, ""));
  return;
}

callback('');
}
},"depart_from":{"form_field":"#departure"},"destination":{"form_field":"#arrival"},"departure_date":{"form_field":"#departureDate"},"return_date":{"form_field":"#returnDate:enabled"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?opodo\\.co\\.uk","name":"Opodo.co.uk","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {

if ($ih('.mmushroomBG.right.r5').length > 0) {
  var match = $ih('.mmushroomBG.right.r5:first').text().match(/([\d\.,]+)\s+GBP/);
  if (match) {
    callback(match[1]);
    return;
  }
}

callback('');


}
},"depart_from":{"url_param":"departureAirportCode"},"destination":{"url_param":"arrivalAirportCode"},"departure_date":{"js":function(html, $ih, callback, scrapeField) {

if ($ih('#departureDay').length > 0 && $ih('#departureMonth').length > 0) {
  callback($ih('#departureMonth').val() + $ih('#departureDay').val())
}


}
},"return_date":{"js":function(html, $ih, callback, scrapeField) {
if ($ih('#tripTypeR').is(':checked')) {
  if ($ih('#returnDay').length > 0 && $ih('#returnMonth').length > 0) {
    callback($ih('#returnMonth').val() + $ih('#returnDay').val());
    return;
  }
}

callback('');






}
},"stops":{"js":function(html, $ih, callback, scrapeField) {

function scrapeLowestStops(selector) {
  var stops = 0;

  $ih(selector).each(function(i, elem) {
    var match = $ih(elem).text().match(/([\d]+)\s+stop/);

    if (match) {
      var parsedStops = parseInt(match[1]);
      if (stops == 0 || parsedStops < stops) stops = parsedStops;
    }
  });

  return stops;
}

callback(scrapeLowestStops('#outer-0 a.stops') + scrapeLowestStops('#inner-01 a.stops'));

}
}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?kayak\\.co\\.uk","name":"Kayak.co.uk","scraper":{"price":{"element_text":".flightresult:first .bookitprice:first","rescrape_for":15},"outbound_stops":{"element_text":".flightresult:first .stops:first","rescrape_for":15},"return_stops":{"element_text":".flightresult:first .stops:eq(1)","rescrape_for":15},"js":function(html, $ih, callback, scrapeField) {

var match = window.location.href.match(/flights.?\/([A-Z]{3})-([A-Z]{3})\/(\d{4}-\d\d-\d\d)\/(\d{4}-\d\d-\d\d)/);

var scrape = {};
if (match) {
  scrape.depart_from = match[1];
  scrape.destination = match[2];
  scrape.departure_date = match[3];
  scrape.return_date = match[4];
}

callback(scrape);
}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^http:\\/\\/travelsearch\\.yahoo\\.co\\.uk\\/","name":"TravelsearchYahoo.co.uk","scraper":{"js":function(html, $ih, callback, scrapeField) {
var scrape = {};

function findAirportCode(location) {
  if ($ih('acronym[title="' + location + '"]:first').length > 0) {
    return $ih('acronym[title="' + location + '"]:first').text();
  }
  return location;
}

var match = $ih('#search_summary h2').text().match(/Flight\s+:\s+(.+)\s+to\s+(.+)/);
if (match) {
  var departFrom = match[1];
  var destination = match[2];

  scrape.depart_from = findAirportCode(departFrom);
  scrape.destination = findAirportCode(destination);
}

match = $ih('#search_summary span:first').text().match(/Dep\.:\s+\w+\s+([\d\/]{8})(?:,\s+Ret\.:\s+\w+\s+([\d\/]{8}))?/);
if (match) {
  scrape.departure_date = match[1];
  if (match[2]) scrape.return_date = match[2];
}

callback(scrape);
}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?expedia\\.com","name":"Expedia.com","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
var totalPrice = $ih('#flightTotalPrice_0');

var price = (totalPrice.length) ? totalPrice : $ih("#flightPriceId0");

callback(price.length ? price.text() : '')

}
},"depart_from":{"form_field":"#inpDepartureLocations0"},"destination":{"form_field":"#inpArrivalLocations0"},"departure_date":{"form_field":"#inpDepartDate1"},"return_date":{"form_field":"#inpReturnDate"},"stops":{"js":function(html, $ih, callback, scrapeField) {
function findStops(element) {
  var stops = 0;
    if (element.length > 0) {
    var match = element.html().match(/(\d+) stops?/i);
    if(match) {
      stops = parseInt(match[1]);
    } else {
      match = element.html().match(/Connect in (.+?\(.+?\))(, .+? \(.+?\))*/i);
      if (match) {
        for (var i = 1; i < match.length; i++) {
          if(match[i]) stops++;
        }
      }
    }
  }
  return stops;
}

var firstFlight = $ih("#Flight0");
var stops = findStops(firstFlight);

var isReturnFlight = html.match(/Choose a complete roundtrip flight/i);
if (isReturnFlight) {
  var returnFlight = $ih("#Flight1");
  stops += findStops(returnFlight);
}

callback(stops);

}
}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?(book)?ryanair\\.com","name":"Ryanair.com","scraper":{"price":{"element_text":".sideRightColB:last"},"departure_date":{"element_text":"#sideFlightsResume1 p:eq(0)"},"return_date":{"element_text":"#sideFlightsResume2 p:eq(0)"},"js":function(html, $ih, callback, scrapeField) {

var scrape = {};


var locations = $ih("#sideFlightsResume1 p:last i");
if (locations && (match = locations.text().match(/^(.+?)\s\W\s(.+?)$/))){
  scrape.depart_from = match[1];
  scrape.destination = match[2];
}

callback(scrape);


}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?britishairways\\.com\\/travel\\/fx","name":"BritishAirways.com","scraper":{"depart_from":{"regex":"\"LKDeparturePoint\"\\] = \"(.+?)\""},"destination":{"regex":"\"LKDestinationPoint\"\\] = \"(.+?)\""},"departure_date":{"regex":"\"LKRequestedDepartureDate\"\\]='(.+?)'"},"return_date":{"regex":"\"LKRequestedReturnDate\"\\]='(.+?)'"},"js":function(html, $ih, callback, scrapeField) {
function parse_price() {
  var outboundPrice = 0;
  var inboundPrice = 0;

  var cell = $ih("#outboundDates td.selected .priceCost");

  if (cell.length) {
    var price = parseFloat(cell.text().replace(/[^0-9.]/g, ''));
    if (price) outboundPrice = price;
  }
  cell = $ih("#inboundDates td.selected .priceCost");
  if (cell.length) {
    var price = parseFloat(cell.text().replace(/[^0-9.]/g, ''));
    if (price) inboundPrice = price;
  }
  return cell.text().substring(0,1) + (inboundPrice + outboundPrice);
}

function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#from");
  if (cell.length) scrape.depart_from = cell.val();

  scrape.destination = "";
  cell = $ih("#to");
  if (cell.length) scrape.destination = cell.val();

  scrape.departure_date = "";
  cell = $ih("#obHolder .calendarSelection input:radio:checked");
  if (cell.length) scrape.departure_date = cell.val();

  scrape.return_date = "";
  cell = $ih("#ibHolder .calendarSelection input:radio:checked");
  if (cell.length) scrape.return_date = cell.val();

  scrape.price = parse_price();

if (!scrape.depart_from && !scrape.destination && !scrape.departure_date) {
var outbound = $ih('#outJourney, #out0Journey').text();
var match;
if (match = outbound.match(/\s*(.+?) - (.+?)\s-/)) {
scrape.depart_from = match[1];
scrape.destination = match[2];
match = outbound.match(/(\d\d? \w{3,9} \d{4})/);
if (match) scrape.departure_date = match[1];
}
var inbound = $ih('#inJourney, #in0Journey').text();
if (match = inbound.match(/(\d\d? \w{3,9} \d{4})/)) {
scrape.return_date = match[1];
}

var outPrice = $ih('#outboundCjs td.price.lowest:first').text().replace(/[^\.\d]/, "");
var inPrice = $ih('#inboundCjs td.price.lowest:first').text().replace(/[^\.\d]/, "");
scrape.price = inPrice == '' ? parseFloat(outPrice) : parseFloat(outPrice) + parseFloat(inPrice);
}

  return scrape;
}

$ih(".calendarSelection input:radio").bind('click',function() {
  if (handler) // to support the refactored/rewritten code
    handler.process(scrape());
  else
    callback(scrape());
})

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(reservations\\.)?jet2\\.com","name":"Jet2.com","scraper":{"price":{"element_text":".runningtotalprice:first","rescrape_for":30}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://book\\.statravel\\.co\\.uk","name":"Statravel.co.uk","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback("UK");
}
},"price":{"regex":"airResultsUITool.filterAirlinesStops\\('',1,8\\);return false;\" href=\"#\">\\$(.+?)<"},"depart_from":{"element_text":"#departureOriginLocationTextRT"},"destination":{"element_text":"#departureDestLocationTextRT"},"departure_date":{"js":function(html, $ih, callback, scrapeField) {
var date = $ih(".flightDetails table tbody tr td:eq(1)").text().match(/..., ..? ... ..../); if(date){callback(date[0]);}else{callback();}
}
},"return_date":{"js":function(html, $ih, callback, scrapeField) {
var date = $ih(".flightDetails table tbody tr:eq(2) td:eq(1)").text().match(/..., ..? ... ..../); if(date){callback(date[0]);}else{callback();}
}
}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?wizzair\\.com","name":"Wizzair.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#from1Select");
  if (cell.length) scrape.depart_from = cell.val();

  scrape.destination = "";
  cell = $ih("#to1Select");
  if (cell.length) scrape.destination = cell.val();

  scrape.departure_date = "";
  var selectedDate = $ih("#AllInclusivePriceSummary_r1_date").text();
  var departureDate = selectedDate !== "" ? selectedDate :
        $ih("#departMonth1Select").val() + $ih("#departDay1Select").val();
  if (departureDate !== "") scrape.departure_date = departureDate;

  scrape.return_date = "";
  selectedDate = $ih("#AllInclusivePriceSummary_r2_date").text();
  var returnDate = selectedDate !== "" ? selectedDate :
        $ih("#departMonth2Select").val() + $ih("#departDay2Select").val();
  if (returnDate !== "") scrape.return_date = returnDate;

  scrape.price = "";
  cell = $ih("#AllInclusivePriceSummary_summary");
  if (cell.length) scrape.price = cell.text();

  return scrape;
}

$ih("input:radio").bind('click', function() {
  if (handler)
    handler.process(scrape());
  else
    callback(scrape());
});

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":null,"name":"Monarch.co.uk","scraper":{"price":{"wait_for":3,"js":function(html, $ih, callback, scrapeField) {
function parsePrice(price) {
  return parseFloat(price.replace(/[^\d\.]/g, ""));
}


var totalPriceText = $ih('.total-price .price').text();
if (totalPriceText != '') {
  var currency = totalPriceText.substring(0, 1);
  var totalPrice = parsePrice(totalPriceText);

  var infantsCost = parsePrice($ih('#breakdownInfants_1').parent().next().html())
          + parsePrice($ih('#breakdownInfants_2').parent().next().html());

  var numberOfPeople = parseInt($ih('#AvailabilitySearchInputSelectView_DropDownListPassengerTypeADT').val())
              + parseInt($ih('#AvailabilitySearchInputSelectView_DropDownListPassengerTypeCHD').val());

  callback( currency + (( totalPrice - infantsCost ) / numberOfPeople) );
} else {
  callback('');
}

}
},"depart_from":{"form_field":"#AvailabilitySearchInputSelectView_DropDownListOrigin1"},"destination":{"form_field":"#AvailabilitySearchInputSelectView_DropDownListDestination1"},"departure_date":{"form_field":"#AvailabilitySearchInputSelectView_TextBoxDate1"},"return_date":{"form_field":"#AvailabilitySearchInputSelectView_TextBoxDate2"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?aerlingus\\.com","name":"AerLingus.com","scraper":{"price":{"element_text":"#displayTotalPrice"},"depart_from":{"regex":"&selectedSourceAirport_1=([A-Z]{3})"},"destination":{"regex":"&selectedDestinationAirport_1=([A-Z]{3})"},"js":function(html, $ih, callback, scrapeField) {
var scrape = {};

if ($ih('h5.center:first').length > 0) {
scrape.departure_date = $ih('h5.center:first').text();
} else {
if ((dDate = $ih('input[name="selectedOutDate"]:checked')).length > 0) {
scrape.departure_date = dDate.val()
}
}

if ($ih('h5.center:eq(1)').length > 0) {
scrape.return_date = $ih('h5.center:eq(1)').text();
} else {
if ((rDate = $ih('input[name="selectedInDate"]:checked')).length > 0) {
scrape.return_date = rDate.val()
}
}

callback(scrape);
}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?delta\\.com\\/booking\\/","name":"Delta.com","scraper":{"price":{"wait_for":10,"element_text":".lowFareBadge .full_fare:first"},"depart_from":{"form_field":"#departureCity\\[0\\]"},"destination":{"form_field":"#destinationCity\\[0\\]"},"departure_date":{"form_field":"#departureDate\\[0\\]"},"return_date":{"form_field":"#departureDate\\[1\\]"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":null,"name":"CharterMonarch.co.uk","scraper":{"depart_from":{"url_param":"departureIATAs"},"destination":{"url_param":"destinationIATAs"},"departure_date":{"url_param":"date"},"return_date":{"url_param":"rtn_date"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(travel.)?\\.priceline\\.com","name":"Priceline.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback("US");
}
},"price":{"sum_values":"#results .pricebox .price:first"},"depart_from":{"form_field":"input[name=\"OFFER/AIRLINE_TICKET/LOCATION[0]/CITY/@NAME\"]:first"},"destination":{"form_field":"input[name=\"OFFER/AIRLINE_TICKET/LOCATION[1]/CITY/@NAME\"]:first"},"departure_date":{"form_field":"#AIRDHTML_DEPARTUREDATE, #AIRDHTML_DEPARTUREDATE_OW"},"return_date":{"form_field":"#AIRDHTML_RETURNDATE"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?americanairlines\\.co\\.uk","name":"AmericanAirlines.co.uk","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
var cell = $ih('.mdblue .sub3:last');
var price = '';
if (cell) {
  price = cell.text().trim().replace(/[^0-9.]+/,'');
}

callback(price);

}
},"depart_from":{"form_field":"#originAirport11111"},"destination":{"form_field":"#departureAirport11111"},"departure_date":{"regex":"name=\"currentDepDate\" value=\"(.+?)\""},"return_date":{"regex":"name=\"currentRetDate\" value=\"(.+?)\""}},"category":"flights_provider","style_override":null},{"region":"UK","regex":null,"name":"Thomson.co.uk","scraper":{"js":function(html, $ih, callback, scrapeField) {
var scrape = {};

scrape.depart_from = "";
try {
  scrape.depart_from = $ih("h3.bookingHead:first").text().split(' to ')[0].match(/\(([A-Z]{3})\)/)[1];
} catch(e) {
}

scrape.destination = "";
try {
  scrape.destination = $ih("h3.bookingHead:first").text().split(' to ')[1].match(/\(([A-Z]{3})\)/)[1];
} catch(e) {
}

scrape.departure_date = "";
var day = $ih("#depDay");
var month = $ih("#depMon");
if (day.length > 0 && month.length > 0) {
  scrape.departure_date = month.val() + "-" + day.val();
}

scrape.return_date = "";
var oneway = $ih("#chkOneWay:checked");
if (!(oneway.length > 0)) {
  var day = $ih("#retDay");
  var month = $ih("#retMon");
  if (day.length > 0 && month.length > 0) {
    scrape.return_date = month.val() + "-" + day.val();
  }
}

scrape.price = "";
cell = $ih(".total .amount:first")
if (cell.length) {
  scrape.price = cell.text().replace(/[^0-9,.]/g, '');
}

callback(scrape);

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?ebookers\\.com","name":"Ebookers.com","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {

if ($ih('.price:first').length > 0) {
  callback($ih('.price:first').text());
  return;
}

callback('');

}
},"depart_from":{"url_param":"ar\\...\\.leaveSlice\\.orig\\.key"},"destination":{"url_param":"ar\\...\\.leaveSlice\\.dest\\.key"},"departure_date":{"url_param":"ar\\...\\.leaveSlice\\.date"},"return_date":{"url_param":"ar\\.rt\\.returnSlice\\.date"},"stops":{"js":function(html, $ih, callback, scrapeField) {

var stops = 0;
$ih('.airResultsCard:first .numStops').each(function(i, elem) {
  if (match = $ih(elem).text().match(/(\d+) stop/)) {
    stops += parseInt(match[1]);
  }
});


callback(stops);


}
}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://www\\.travelocity\\.co\\.uk","name":"Travelocity.co.uk","scraper":{"price":{"element_text":"#price-range-min"},"depart_from":{"form_field":"input[name=\"ac_departure_airport\"]:first"},"destination":{"form_field":"input[name=\"ac_arrival_airport\"]:first"},"js":function(html, $ih, callback, scrapeField) {
var scrape = {};

var departureMonth = $ih("input[name='depart_month']:first").val();
var departureDay =  $ih("input[name='depart_day']:first").val();
if (departureMonth && departureDay) scrape.departure_date = departureMonth + "-" + departureDay;

var returnMonth = $ih("input[name='return_month']:first").val();
var returnDay = $ih("input[name='return_day']:first").val();
if (returnMonth && returnDay) scrape.return_date = returnMonth + "-" + returnDay;

callback(scrape);
}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://book\\.jetblue\\.com","name":"JetBlue.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#ulistRouteEndPoints_0 li:first");
  if (cell.length) scrape.depart_from = cell.text();

  scrape.destination = "";
  cell = $ih("#ulistRouteEndPoints_0 li:last");
  if (cell.length) scrape.destination = cell.text();

  if ($ih('.active-day h3').length > 0) {
  scrape.departure_date = $ih('.active-day h3:first').text();
  }

  if ($ih('.active-day h3').length == 2) {
  scrape.return_date = $ih('.active-day h3:last').text();
  }

  scrape.price = "";
  var outboundPrice = 0;
  var inboundPrice = 0;
  $ih(".outboundFlightRow .orange-txt-color span:first").text().replace("$","")
  cell = $ih(".outboundFlightRow .orange-txt-color span:first");
  if (cell.length) {
    outboundPrice = parseInt(cell.text().replace('$',''));
  }
  cell = $ih(".inboundFlightRow .orange-txt-color span:first");
  if (cell.length) {
    inboundPrice = parseInt(cell.text().replace('$',''));
  }
  scrape.price = outboundPrice + inboundPrice;
  if (scrape.price == 0) scrape.price = 0;

  return scrape;
}

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotwire\\.com/air","name":"Hotwire.com","scraper":{"price":{"wait_for":20,"js":function(html, $ih, callback, scrapeField) {

var priceText = $ih('.result .price:first strong').text();
if (priceText !== "") {
  var price = parseInt(priceText.replace(/[^\d]/g, "")) / 100;
  callback(price);
  return;
}

callback('');


}
},"depart_from":{"form_field":"input[name=\"origCity\"]:first"},"destination":{"form_field":"input[name=\"destinationCity\"]:first"},"outbound_stops":{"wait_for":20,"min_value":".origin .stopsAndDuration strong"},"return_stops":{"wait_for":20,"min_value":".destination .stopsAndDuration strong"},"departure_date":{"regex":"u3=(.+?)&"},"return_date":{"regex":"u4=(.+?)&"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://travel\\.united\\.com","name":"United.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih(".fromTo:eq(0)");
  if (cell.length) scrape.depart_from = cell.text();

  scrape.destination = "";
  if ($ih(".fromTo").length == 1) {
    cell = $ih(".fromTo");
    scrape.destination = cell.text().split(/\sto\s/)[1];
  } else {
    cell = $ih(".fromTo:eq(1)");
    if (cell.length) scrape.destination = cell.text();
  }

  scrape.departure_date = "";
  cell = $ih(".fsndate:first");
  if (cell.length) {
    scrape.departure_date = cell.text();
  }

  scrape.return_date = "";
  cell = $ih(".fsndate");
  if (cell.length == 2) {
    scrape.return_date = $ih(".fsndate:last").text();
  }

  cell = $ih("#anyTimeLow");
  if (cell.length) scrape.price = cell.text();

  return scrape;
}

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(ww.\\.)?thomascook\\.com.+?flight_only\\/results","name":"ThomasCook.com","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
var text = $ih('.plusMinusDayPrice').text();
callback(text ? parseFloat(text.replace(/[^\d.]/g, '')) : '');
}
},"depart_from":{"js":function(html, $ih, callback, scrapeField) {
var text = $ih('.titleBarText').text();
var match;
if (text) match = text.replace(/\s+/g, ' ').match(/from (.+) to/);
callback(match ? match[1] : '')
}
},"destination":{"js":function(html, $ih, callback, scrapeField) {
var text = $ih('.titleBarText').text();
var match;
if (text) match = text.replace(/\s+/g, ' ').match(/to (.+)$/);
callback(match ? match[1] : '')
}
},"return_date":{"js":function(html, $ih, callback, scrapeField) {
var date = $ih('#retDateField input').val();
if (date && (!$ih('#retDateField input').attr('disabled')))
  callback(date)
else
  callback('')
}
}},"category":"flights_provider","style_override":null},{"region":null,"regex":"^http:\\/\\/(www\\.)?bing\\.com\\/search","name":"Bing.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
var isFlightSearch = function(query) {
  var match = query.match(/(flights?|fl(y|ights?)(\s+(from|to).+))/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isFlightSearch(query)) return;

var res = {};

var match = query.match(/from\s+(.+?)(?:\s+to|$)/);
if (match) res.depart_from = match[1];

var match = query.match(/to\s+(.+)/);
if (match) res.destination = match[1];

res.region = scrapeField(html, '<span class="mktInd"><a[^>]+>([^<]+)</a>') == 'United Kingdom' ? 'UK' : 'US';

callback(res);
}
},"category":"flights_provider","style_override":"#sw_im { \n  top: <<0>>px !important;\n}\nbody {\n  background-position: 0px <<-76>>px !important;\n}\n"},{"region":"US","regex":"^https?://tickets\\.airtran\\.com","name":"Airtran.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  var cell = $ih("#flightmatrix h3");
  if (cell.length) {
    try {
      scrape.depart_from = $ih("#flightmatrix h3").text().split(":")[1].split(" to ")[0].split(",")[0];
      scrape.destination = $ih("#flightmatrix h3").text().split(":")[1].split(" to ")[1].split(",")[0];
    } catch(e) {
    }
  }

  scrape.departure_date = "";
  var day = $ih("#ATAvailabilitySearchInputSelectView_DropDownListMarketDay1");
  if (day) day = day.val();
  var month = $ih("#ATAvailabilitySearchInputSelectView_DropDownListMarketMonth1");
  if (month) month = month.val();
  scrape.departure_date = month + "-" + day;

  scrape.return_date = "";
  var day = $ih("#ATAvailabilitySearchInputSelectView_DropDownListMarketDay2:visible");
  if (day) day = day.val();
  var month = $ih("#ATAvailabilitySearchInputSelectView_DropDownListMarketMonth2:visible");
  if (month) month = month.val();
  if (month && day) {
    scrape.return_date = month + "-" + day;
  }

  return scrape;
}

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?orbitz\\.com","name":"Orbitz.com","scraper":{"price":{"regex":"\"DCSext.lps\" content=\"(.+?)\""},"depart_from":{"url_param":"ar.rt.leaveSlice.orig.key"},"destination":{"url_param":"ar.rt.leaveSlice.dest.key"},"outbound_stops":{"min_value":".result .resultSlice:nth-child(2) .resultLeg:first-child .numStops"},"return_stops":{"min_value":".result .resultSlice:nth-child(3) .resultLeg:first-child .numStops"},"departure_date":{"url_param":"ar.rt.leaveSlice.date"},"return_date":{"url_param":"ar.rt.returnSlice.date"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?southwest\\.com","name":"Southwest.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function getLowestPriceAndStopsFromTable(selector) {
  var lowestPrice;
  var lowestRow;
  $ih(selector + " .bugTableRow").each(function(i, rowElem) {
    var row = $ih(rowElem);
    row.find(".product_price").each(function(j, priceElem) {
      var price = parseInt($ih(priceElem).text().replace(/\$/, ""));
      if (!lowestPrice || price < lowestPrice) {
        lowestPrice = price;
        lowestRow = row;
      }
    });
  });

  if (lowestPrice) {
    var res = {};
    res.price = lowestPrice;
    res.stops = parseInt(lowestRow.find(".routing_column .bugText a").text().replace(/ stops?/, "")) || 0;
    return res;
  }
  return null;
}

var res = {};

res.price = $ih("#priceItinerary_grandTotal_allPaxTypes").text();
var stops = 0;
$ih(".stops").each(function(i, elem) {
  var match = $ih(elem).text().match(/\((\d+) stops?/);
  stops += parseInt(match[1]);
});
res.stops = stops;

if (!res.price) {
  var outbound = getLowestPriceAndStopsFromTable("#faresOutbound");
  var inbound = getLowestPriceAndStopsFromTable("#faresReturn");

  if (outbound) {
    res.price = outbound.price;
    res.stops = outbound.stops;
  }
  if (inbound) {
    res.price += inbound.price;
    res.stops += inbound.stops;
  }
}

var match = $ih("#originAirport option:selected").text().match(/.+? - ([A-Z]{3})/);
if (match) {
  res.depart_from = match[1];
}
match = $ih("#destinationAirport option:selected").text().match(/.+? - ([A-Z]{3})/);
if (match) {
  res.destination = match[1];
}

var departureDate = $ih("#outboundDate").val();
if (departureDate) {
  res.departure_date = departureDate;
}

var isReturnFlight = $ih("#returnAirport_displayed").val() != 'None';
var returnDate = $ih("#returnDate").val();
if (isReturnFlight && returnDate) {
  res.return_date = returnDate;
}

callback(res);
}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?aa\\.com\\/","name":"AmericanAirlines.com","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
var cell = $ih('.mdblue .sub3:last');
var price = '';
if (cell) {
  price = cell.text().trim().replace(/[^0-9.]+/,'');
}

callback(price);

}
},"depart_from":{"form_field":"#originAirport11111"},"destination":{"form_field":"#departureAirport11111"},"departure_date":{"regex":"name=\"currentDepDate\" value=\"(.+?)\""},"return_date":{"regex":"name=\"currentRetDate\" value=\"(.+?)\""}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?edreams\\.co\\.uk","name":"Edreams.co.uk","scraper":{"price":{"element_text":".singleItinerayPrice:first"},"depart_from":{"form_field":"#departureLocation"},"destination":{"form_field":"#arrivalLocation"},"departure_date":{"form_field":"#departureDate"},"return_date":{"form_field":"#returnDate"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?flybe\\.com","name":"Flybe.com","scraper":{"price":{"element_text":"#totalValue","rescrape_for":"120"},"depart_from":{"form_field":"#departureSelect"},"destination":{"form_field":"#destinationSelect"},"departure_date":{"element_text":"#sector1 #date"},"return_date":{"element_text":"#sector2 #date"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://shopping\\.usairways\\.com","name":"UsAirways.com","scraper":{"price":{"sum_values":".itinlist:first td.bestmatchoff:first, .itinlist:eq(1) td.bestmatchoff:first"},"depart_from":{"form_field":"#ctl00_phMain_ModifySearchControl_TravelMarketsEntryControl_TravelCitiesEntryControl_OriginAirportInput_textbox"},"destination":{"form_field":"#ctl00_phMain_ModifySearchControl_TravelMarketsEntryControl_TravelCitiesEntryControl_DestinationAirportInput_textbox"},"departure_date":{"form_field":"#ctl00_phMain_ModifySearchControl_TravelMarketsEntryControl_TravelDatesEntryControl_OutboundDateInput_SelectedDate"},"return_date":{"form_field":"#ctl00_phMain_ModifySearchControl_TravelMarketsEntryControl_TravelDatesEntryControl_ReturnDateInput_SelectedDate"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://book\\.statravel\\.com","name":"Statravel.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback("US");
}
},"price":{"regex":"airResultsUITool.filterAirlinesStops\\('',1,8\\);return false;\" href=\"#\">\\$(.+?)<"},"depart_from":{"element_text":"#departureOriginLocationTextRT"},"destination":{"element_text":"#departureDestLocationTextRT"},"departure_date":{"js":function(html, $ih, callback, scrapeField) {
var date = $ih(".flightDetails table tbody tr td:eq(1)").text().match(/..., ..? ... ..../); if(date){callback(date[0]);}else{callback();}
}
},"return_date":{"js":function(html, $ih, callback, scrapeField) {
var date = $ih(".flightDetails table tbody tr:eq(2) td:eq(1)").text().match(/..., ..? ... ..../); if(date){callback(date[0]);}else{callback();}
}
}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?skyscanner\\.net","name":"Skyscanner.net","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
var priceTextElements = $ih('.price_text');
if (priceTextElements.length > 0) {
  var price;
  var currency = "";
  if (currMatch = priceTextElements.text().match(/([\$])/)) {
    currency = currMatch[1];
  }

  priceTextElements.each(function(i, elem) {
    var scrapedPrice = $ih(elem).text().replace(/[^\d\.]/, "");
    price = !price || price > scrapedPrice ? scrapedPrice : price;
  });

  callback( currency + price );
} else {
  callback('');
}
}
},"depart_from":{"form_field":"#sc_fromPlace"},"destination":{"form_field":"#sc_toPlace"},"departure_date":{"form_field":"#sc_departDate"},"return_date":{"js":function(html, $ih, callback, scrapeField) {
var returnDate = $ih('#sc_returnDate').val();

if (returnDate !== "" && returnDate !== "One way") {
  callback(returnDate);
  return;
}

callback('');

}
},"stops":{"js":function(html, $ih, callback, scrapeField) {
var stopsText = $ih('.day_resultslist .stops:first').text();

if (stopsText !== "") {
  if (match = stopsText.match(/(\d+)/)) {
    callback(match[1]);
    return;
  }
}

callback('');


}
}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://travel\\.travelocity\\.com","name":"Travelocity.com","scraper":{"price":{"element_text":".lowest-fare:first a","rescrape_for":15},"depart_from":{"form_field":"input[name=\"leavingFrom\"]:first"},"destination":{"form_field":"input[name=\"goingTo\"]:first"},"departure_date":{"form_field":"#fo-fromdate, #fromdateexact"},"return_date":{"form_field":"#fo-todate, #todateexact"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?easyjet\\.com","name":"Easyjet.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrapeDate(dayInput, monthYearInput) {
  var day = $ih(dayInput).val();
  var monthYear = $ih(monthYearInput).val();
  if (day && monthYear) {
    return (day.toString().length == 1 ? '0' + day : day) + monthYear;
  }
  return null;
}

function scrapePrice(container) {
  var price = $ih("div." + container + " li.alt-lowest:visible:first span.priceSmaller").text().replace(/[^\d]/, "");
  return price ? parseInt(price) / 100 : null;
}

function scrape() {
  var scrape = {};

  if (departFrom = $ih("#acOriginAirport").val()) {
    scrape.depart_from = departFrom;
  }

  if (destination = $ih("#acDestinationAirport").val()) {
    scrape.destination = destination;
  }

  if (outDate = scrapeDate('#oDay', '#oMonYear')) {
    scrape.departure_date = outDate;
  }

  if($ih("#chkIsReturn:checked").length > 0) {
    if (returnDate = scrapeDate('#rDay', '#rMonYear')) {
      scrape.return_date = returnDate;
    }
  }

  var price, returnPrice;
  if (price = scrapePrice('OutboundDaySliderContainer')) {
    if (returnPrice = scrapePrice('ReturnDaySliderContainer')) {
      price += returnPrice;
    }
    if (match = $ih('span.priceSmaller:first').text().match(/([\u20ac\u00a3\$])/))  {
      price = match[1] + price;
    }
    scrape.price = price;
  }
  return scrape;
}

$ih("#ChangeSearchButton").bind('click', function() {
  if (handler)
    handler.process(scrape());
  else
    callback(scrape());
});

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?klm\\.com","name":"Klm.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#appFrame").contents().find("select[name='connections[0].originCode']");
  if (cell.length) scrape.depart_from = cell.val();

  scrape.destination = "";
  cell = $ih("#appFrame").contents().find("input[name='connections[0].destination']");
  if (cell.length) scrape.destination = cell.val();

  scrape.departure_date = "";
  cell = $ih("#appFrame").contents().find("#ebt-departure-date");
  if (cell.length) scrape.departure_date = cell.val();

  scrape.return_date = "";
  cell = $ih("#appFrame").contents().find("#ebt-return-date");
  if (cell.length) scrape.return_date = cell.val();

  scrape.price = "";
  cell = $ih("#appFrame").contents().find("#ebt-selected-flight label");
  if (cell.length) scrape.price = cell.text();

  return scrape;
}

$ih("#appFrame").bind('load', function() {
  callback(scrape());
});

callback(scrape());
}
},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?cheapoair\\.com","name":"Cheapoair.com","scraper":{"price":{"min_value":"a.orange span.htloptiontxt"},"depart_from":{"form_field":"#ctl07_ctl02_ctl00_FlightModifySearch_OutBound_From"},"destination":{"form_field":"#ctl07_ctl02_ctl00_FlightModifySearch_OutBound_To"},"departure_date":{"form_field":"#ctl07_ctl02_ctl00_FlightModifySearch_OutBound_Date"},"return_date":{"form_field":"#ctl07_ctl02_ctl00_FlightModifySearch_Return_Date:visible"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(ach\\.)?travel\\.yahoo\\.net","name":"TravelYahoo.net","scraper":{"price":{"element_text":".tfPrice:first .perPerson"},"depart_from":{"form_field":"input[name=\"leavingFrom\"]:first"},"destination":{"form_field":"input[name=\"goingTo\"]:first"},"departure_date":{"form_field":"#fromdateexact"},"return_date":{"form_field":"#todateexact"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?cheaptickets\\.com/App","name":"CheapTickets.com","scraper":{"price":{"regex":"\"DCSext.lps\" content=\"(.+?)\""},"depart_from":{"regex":"DCSext.o\" content=\"(.+?)\""},"destination":{"regex":"DCSext.d\" content=\"(.+?)\""},"departure_date":{"regex":"DCSext.dd\" content=\"(.+?)\""},"return_date":{"regex":"DCSext.rd\" content=\"(.+?)\""},"stops":{"js":function(html, $ih, callback, scrapeField) {
var outboundStops = null;
var returnStops = null;
var stops = null;

$ih('.nextSlice').each(function(i, elem) {
  var st = $ih(elem).find('.legAlerts .alert:contains(Change)').length;
  outboundStops = (outboundStops == null) ? st : Math.min(outboundStops, st);
});

$ih('div:not(.nextSlice) > .slice').each(function(i, elem) {
  var st = $ih(elem).find('.legAlerts .alert:contains(Change)').length;
  returnStops = (returnStops == null) ? st : Math.min(outboundStops, st);
});

if (outboundStops != null || returnStops != null) {
  stops = (outboundStops || 0) + (returnStops || 0);
}

callback( (stops == null) ? '' : stops );

}
}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?:\\/\\/(www\\.)?google\\.com\\/(?:#|search|webhp)","name":"Google.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
var isFlightSearch = function(query) {
  var match = query.match(/(flights?|fl(y|ights?)(\s+(from|to|search).+))/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isFlightSearch(query)) return;

var res = {};

var match = query.match(/from\s+(.+?)(?:\s+to|$)/);
if (match) res.depart_from = match[1];

var match = query.match(/to\s+(.+)/);
if (match) res.destination = match[1];

var match = query.match(/search/);
if (match) res.destination = " ";

callback(res);
}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?lastminute\\.com","name":"Lastminute.com","scraper":{"price":{"element_text":".product-price:first .number"},"depart_from":{"regex":"<li class=\"destination\">[\\s\\S].+?(\\w.+?) -"},"destination":{"regex":"<li class=\"destination\">[\\s\\S].+?\\w.+? - (.+?)\\s"},"departure_date":{"element_text":"#search_recap .dates .date:eq(0)"},"return_date":{"element_text":"#search_recap .dates .date:eq(1)"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^http:\\/\\/(www\\.)?bing\\.com\\/travel","name":"Bing.com/travel","scraper":{"price":{"element_text":".resultsTable td.price:first","rescrape_for":20},"depart_from":{"form_field":"#orig1Text"},"destination":{"form_field":"#dest1Text"},"outbound_stops":{"min_value":".row1 .stops","rescrape_for":20},"return_stops":{"min_value":".row2 .stops","rescrape_for":20},"departure_date":{"form_field":"#leave1"},"return_date":{"form_field":"#return1"}},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?:\\/\\/(www\\.)?google\\.co\\.uk\\/(?:#|search|webhp)","name":"Google.co.uk","scraper":{"js":function(html, $ih, callback, scrapeField) {
var isFlightSearch = function(query) {
  var match = query.match(/(flights?|fl(y|ights?)(\s+(from|to).+))/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isFlightSearch(query)) return;

var res = {};

var match = query.match(/from\s+(.+?)(?:\s+to|$)/);
if (match) res.depart_from = match[1];

var match = query.match(/to\s+(.+)/);
if (match) res.destination = match[1];

callback(res);
}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?travelsupermarket\\.com","name":"TravelSupermarket.com","scraper":{"js":function(html, $ih, callback, scrapeField) {
function scrape() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih(".search-term:eq(2)");
  if (cell.length) scrape.depart_from = cell.val();

  scrape.destination = "";
  cell = $ih(".search-term:eq(2)");
  if (cell.length) scrape.destination = cell.val();

  scrape.departure_date = "";
  cell = $ih(".search-term:eq(3)");
  if (cell.length) scrape.departure_date = cell.val();

  scrape.return_date = "";
  cell = $ih("#ucFlightResultsSearch_rdpReturnDate_dateInput_text");
  if (cell.length) scrape.return_date = cell.val();

  var priceLink = $ih('.priceLink:first');
  scrape.price = 0;
  if (priceLink.length > 0) {
    scrape.price = priceLink.text();
  }

  scrape.stops = 0;
  var outboundElement = $ih(".outbound-leg:first").text();
  if (match = outboundElement.match(/(\d+) stop/i)) {
    scrape.stops += parseInt(match[1]);
  }
  var inboundElement = $ih(".inbound-leg:first").text();
  if (match = inboundElement.match(/(\d+) stop/i)) {
    scrape.stops += parseInt(match[1]);
  }

  return scrape;
}

callback(scrape());
}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://booktrip\\.flybmi\\.com","name":"FlyBmi.com","scraper":{"price":{"element_text":".TotalValue:first","rescrape_for":"120"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://(www\\.)?kayak\\.com\\/.?flights","name":"Kayak.com","scraper":{"price":{"element_text":".flightresult:first .bookitprice:first","rescrape_for":15},"outbound_stops":{"element_text":".flightresult:first .stops:first","rescrape_for":15},"return_stops":{"element_text":".flightresult:first .stops:eq(1)","rescrape_for":15},"js":function(html, $ih, callback, scrapeField) {

var match = window.location.href.match(/flights\/([A-Z]{3})-([A-Z]{3})\/(\d{4}-\d\d-\d\d)\/(\d{4}-\d\d-\d\d)/);

var scrape = {};
if (match) {
  scrape.depart_from = match[1];
  scrape.destination = match[2];
  scrape.departure_date = match[3];
  scrape.return_date = match[4];
}

callback(scrape);
}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?virgin-atlantic\\.com/en/gb.+?checkProcessingStatus","name":"VirginAtlantic.com/UK/Flights","scraper":{"price":{"js":function(html, $ih, callback, scrapeField) {
if ($ih('.total').length > 0) {
  callback($ih('.total:first').text().replace("Total fare", ""));
  return;
} else if ($ih('.priceCol strong').length > 0) {
  callback($ih('.priceCol strong:first').text().replace(/per adult.*/, ""));
  return;
}

callback('');
}
},"depart_from":{"form_field":"#departure"},"destination":{"form_field":"#arrival"},"departure_date":{"form_field":"#departureDate"},"return_date":{"form_field":"#returnDate:enabled"}},"category":"flights_provider","style_override":null},{"region":"US","regex":"^https?://www\\.continental\\.com","name":"Continental.com","scraper":{"departure_date":{"form_field":"#ctl00_ContentInfo_Results_SearchBox_DateTimeCabin1_Depdate_txtDptDate"},"return_date":{"form_field":"#ctl00_ContentInfo_Results_SearchBox_ReturndateTimeCabin_Retdate_txtRetDate"},"js":function(html, $ih, callback, scrapeField) {
function scrapeFlight() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#ctl00_ContentInfo_Results_availSummary1_lblDepartCity");
  if (cell.length) scrape.depart_from = cell.text();

  scrape.destination = "";
  cell = $ih("#ctl00_ContentInfo_Results_availSummary1_lblArriveCity");
  if (cell.length) scrape.destination = cell.text();

  scrape.departure_date = "";
  cell = $ih("#ctl00_ContentInfo_Results_availSummary1_lblDepartDate");
  if (cell.length) {
    scrape.departure_date = cell.text();
  }

  return scrape;
}

function scrapeReview() {
  var scrape = {};

  scrape.depart_from = "";
  var cell = $ih("#mainContent table table tr td.tdSegmentCell:eq(0)");
  if (cell.length) {
    var matches = cell.text().match(/\(([A-Z]{3}) /);
    if (matches) scrape.depart_from = matches[1];
  }

  scrape.destination = "";
  cell = $ih("#mainContent table table tr td.tdSegmentCell:eq(1)");
  if (cell.length) {
    var matches = cell.text().match(/\(([A-Z]{3}) /);
    if (matches) scrape.destination = matches[1];
  }


  return scrape;
}

function scrape() {
  var cell = $ih("#ctl00_ContentPageHeading_PageHeader1_lblPageHeading");
  if (cell.length) {
    var title = cell.text();
    if (title == "Review Trip Itinerary") {
      return scrapeReview();
    }
  }
  return scrapeFlight();
}

callback(scrape());

}
},"category":"flights_provider","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?expedia\\.co\\.uk","name":"Expedia.co.uk","scraper":{"price":{"element_text":"#flightPriceId0"},"depart_from":{"form_field":"#inpDepartureLocations0"},"destination":{"form_field":"#inpArrivalLocations0"},"departure_date":{"form_field":"#inpDepartDate1"},"return_date":{"form_field":"#inpReturnDate"}},"category":"flights_provider","style_override":null}]};
com.forward.invisiblehand.hotels.scrapers = {"regions":{"uk":{"name":"United Kingdom","code":"UK","server":"uk.lb.invisiblehand.co.uk","currency":{"name":"GBP","symbol":"\u00a3"}},"de":{"name":"Germany","code":"DE","server":"de.lb.invisiblehand.co.uk","currency":{"name":"EUR","symbol":"\u20ac"}},"ca":{"name":"Canada","code":"CA","server":"ca.lb.invisiblehand.co.uk","currency":{"name":"CAD","symbol":"$"}},"us":{"name":"United States","code":"US","server":"us.lb.invisiblehand.co.uk","currency":{"name":"USD","symbol":"$"}}},"operators":[{"region":"DE","regex":"^https?://(de\\.)?hotels\\.com","name":"Hotels.de","scraper":{"checkout_date":{"form_field":"input[name=\"departureDate\"]:first"},"location":{"form_field":"input[name=\"destinationName\"]:first, input[name=\"hotelCity\"]:first"},"hotel":{"form_field":"input[name=\"hotelName\"]:first"},"checkin_date":{"form_field":"input[name=\"arrivalDate\"]:first"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
 var c = $ih(".flags").children().first().attr('class'); if(c == 'DEU'){callback('%d.%m.%Y');}else{callback('%m/%d/%y');}
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?opodo\\.co.uk","name":"Opodo.co.uk","scraper":{"checkout_date":{"url_param":"checkOutDate"},"location":{"url_param":"destinationName"},"checkin_date":{"url_param":"checkInDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y%m%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(search\\.)?yahoo\\.com\\/","name":"Yahoo.com","scraper":{"checkout_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+31);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+30);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"js":function(html, $ih, callback, scrapeField) {
var isHotelSearch = function(query) {
  var match = query.match(/(hotels?|hostels?)(\s+(in|near).+)/i);
  return !!match;
}

var query = $ih("input[name='p']").val();
if (!isHotelSearch(query)) return;

var res = {};

var match = query.match(/near\s+(.+)/);
if (match) res.location = match[1];

var match = query.match(/in\s+(.+)/);
if (match) res.location = match[1];

callback(res);
}
},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?orbitz\\.com","name":"Orbitz.com","scraper":{"checkout_date":{"url_param":"hotel.chkout"},"location":{"url_param":"hotel.keyword.key"},"checkin_date":{"url_param":"hotel.chkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?octopustravel\\.com","name":"OctopusTravel.com","scraper":{"checkout_date":{"form_field":"#checkOutDate"},"location":{"form_field":"#searchName"},"hotel":{"regex":"h1>(.+?)<"},"checkin_date":{"form_field":"#checkInDate"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?venere\\.com","name":"Venere.com","scraper":{"checkout_date":{"form_field":"#checkout"},"location":{"regex":"\"geo_name\":\"(.+?)\""},"checkin_date":{"form_field":"#checkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?expedia\\.co\\.uk\\/(Hotel-Search|.+?Hotel-Information)","name":"Expedia.co.uk","scraper":{"checkout_date":{"form_field":"#inpEndDate, #hot-check-out, #hot-info-check-out"},"location":{"form_field":"#inpCity, #hot-sear-near, #hot-info-sear-near, #inpSearchNear"},"hotel":{"js":function(html, $ih, callback, scrapeField) {
 callback($ih('meta[property="og:title"]').attr('content'))
}
},"checkin_date":{"form_field":"#inpStartDate, #hot-check-in, #hot-info-check-in"},"rooms":{"form_field":"#hot-info-num-rooms, #inpRoom"},"adults":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumAdultsInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
},"children":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumChildrenInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?trivago\\.com\\/region.php","name":"Trivago.com","scraper":{"checkout_date":{"form_field":"#dealform_hotel_to_date"},"location":{"url_param":"query_path_name"},"hotel":{"form_field":"#name"},"checkin_date":{"form_field":"#dealform_hotel_from_date"}},"category":"hotel","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?tripadvisor\\.de\\/Hotel_Review","name":"TripAdvisor.de","scraper":{"checkout_date":{"regex":"checkOut\".+?value=\"(.+?\\/.+?\\/.{4})\""},"location":{"regex":"\"location fl\">[\\S\\s]+?(.+?)<\\/div>[\\S\\s]+?<\\/div>[\\S\\s]<\\/li>[\\S\\s]<\\/ul>"},"hotel":{"regex":"h1 id=\"HEADING\" property=\"v:name\">[\\S\\s](.+?)<"},"checkin_date":{"regex":"checkIn\".+?value=\"(.+?\\/.+?\\/.{4})\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m.%d.%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?roomorama\\.com","name":"Roomorama.com","scraper":{"checkout_date":{"form_field":"#check_out"},"location":{"regex":"city\">(.+)<"},"checkin_date":{"form_field":"#check_in"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(globaltrips\\.|www\\.)?lastminute\\.com","name":"LastMinute.com","scraper":{"checkout_date":{"url_param":"checkOutDate"},"location":{"url_param":"city"},"hotel":{"regex":"\"hotelName\":\"(.+?) -"},"checkin_date":{"url_param":"checkInDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?expedia\\.de\\/(Hotel-Search|.+?Hotel-Information)","name":"Expedia.de","scraper":{"checkout_date":{"form_field":"#inpEndDate, #hot-check-out, #hot-info-check-out"},"location":{"form_field":"#inpCity, #hot-sear-near, #hot-info-sear-near, #inpSearchNear"},"hotel":{"js":function(html, $ih, callback, scrapeField) {
 callback($ih('meta[property="og:title"]').attr('content'))
}
},"checkin_date":{"form_field":"#inpStartDate, #hot-check-in, #hot-info-check-in"},"rooms":{"form_field":"#hot-info-num-rooms, #inpRoom"},"adults":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumAdultsInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d.%m.%Y')
}
},"children":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumChildrenInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?google\\.co.uk\\/","name":"Google.co.uk","scraper":{"checkout_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+31);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+30);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"js":function(html, $ih, callback, scrapeField) {
var isHotelSearch = function(query) {
  var match = query.match(/(hotels?|hostels?)(\s+(in|near).+)/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isHotelSearch(query)) return;

var res = {};

var match = query.match(/near\s+(.+)/);
if (match) res.location = match[1];

var match = query.match(/in\s+(.+)/);
if (match) res.location = match[1];

callback(res);
}
},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?bing\\.com\\/","name":"Bing.com","scraper":{"checkout_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+31);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+30);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"js":function(html, $ih, callback, scrapeField) {
var isHotelSearch = function(query) {
  var match = query.match(/(hotels?|hostels?)(\s+(in|near).+)/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isHotelSearch(query)) return;

var res = {};

var match = query.match(/near\s+(.+)/);
if (match) res.location = match[1];

var match = query.match(/in\s+(.+)/);
if (match) res.location = match[1];

callback(res);
}
},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(from\\.)?bookit\\.com","name":"BookIt.com","scraper":{"checkout_date":{"form_field":"#td"},"location":{"form_field":"#toAreaLocation"},"hotel":{"form_field":"input[name=\"hotelName\"]:first"},"checkin_date":{"form_field":"#fd"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hostels\\.com","name":"Hostels.com","scraper":{"checkout_date":{"regex":"prop14\":\"(.+?)\""},"location":{"regex":"prop3\":\"(.+?)\""},"checkin_date":{"regex":"prop13\":\"(.+?)\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?cheaptickets\\.com","name":"CheapTickets.com","scraper":{"checkout_date":{"url_param":"hotel.chkout"},"location":{"url_param":"hotel.keyword.key"},"hotel":{"regex":"h1>(.+?)<"},"checkin_date":{"url_param":"hotel.chkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?trivago\\.co.uk\\/region.php","name":"Trivago.co.uk","scraper":{"checkout_date":{"form_field":"#dealform_hotel_to_date"},"location":{"url_param":"query_path_name"},"hotel":{"form_field":"#name"},"checkin_date":{"form_field":"#dealform_hotel_from_date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(inventory\\.|www\\.)?cheaphotels\\.co.uk","name":"CheapHotels.co.uk","scraper":{"checkout_date":{"regex":"s.eVar10=\"HOT | (.+?)\""},"location":{"form_field":"#enter-city"},"checkin_date":{"regex":"s.eVar10=\"HOT | (.+?)\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y%m%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?booking\\.com\\/(hotel|searchresults)","name":"Booking.com","scraper":{"checkout_date":{"regex":"checkout_date = '(.{4}-.{2}-.{2})'"},"location":{"form_field":"#destination"},"hotel":{"form_field":"#hp_hotel_name"},"checkin_date":{"regex":"checkin_date = '(.{4}-.{2}-.{2})'"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?laterooms\\.com","name":"LateRooms.com","scraper":{"checkout_date":{"regex":"dout=(.{4}-.{2}-.{2})"},"location":{"form_field":"#txtSearch"},"checkin_date":{"regex":"din=(.{4}-.{2}-.{2})"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotelclub\\.com","name":"HotelClub.com","scraper":{"checkout_date":{"url_param":"hotel.chkout"},"location":{"url_param":"hotel.keyword.key"},"checkin_date":{"url_param":"hotel.chkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%y')
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?travelpack\\.com","name":"Travelpack.com","scraper":{"checkout_date":{"form_field":"#Depart"},"location":{"regex":"name=\"Location\" value=\"(.+?) \\["},"checkin_date":{"form_field":"#Arrival"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?google\\.com\\/","name":"Google.com","scraper":{"checkout_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+31);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
var myDate=new Date();myDate.setDate(myDate.getDate()+30);myDate.setMonth(myDate.getMonth()+1);callback((myDate.getMonth() < 10 ? '0' : '' ) + myDate.getMonth() + '/' + (myDate.getDate() < 10 ? '0' : '' ) + myDate.getDate() + '/' + myDate.getFullYear());
}
},"js":function(html, $ih, callback, scrapeField) {
var isHotelSearch = function(query) {
  var match = query.match(/(hotels?|hostels?)(\s+(in|near).+)/i);
  return !!match;
}

var query = $ih("input[name='q']").val();
if (!isHotelSearch(query)) return;

var res = {};

var match = query.match(/near\s+(.+)/);
if (match) res.location = match[1];

var match = query.match(/in\s+(.+)/);
if (match) res.location = match[1];

callback(res);
}
},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(travel.\\.)?priceline\\.com","name":"Priceline.com","scraper":{"checkout_date":{"form_field":"#tmpCheckOutDate"},"location":{"form_field":"#destName"},"checkin_date":{"form_field":"#tmpCheckInDate","rescape_for":10}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?skoosh\\.co.uk","name":"SkooshCoUK","scraper":{"checkout_date":{"form_field":"input[name=\"check_out_date\"]:first"},"location":{"form_field":"#city_specified, h3:eq(0)"},"hotel":{"regex":"<h1 itemprop=\"name\">(.+?)<\\/h1>"},"checkin_date":{"form_field":"input[name=\"check_in_date\"]:first"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?accorhotels\\.com","name":"AccorHotels.com","scraper":{"checkout_date":{"form_field":"#search-dateOut-new"},"location":{"form_field":"#search-destination-new"},"checkin_date":{"form_field":"#search-dateIn-new"}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?tripadvisor\\.co.uk\\/Hotel_Review","name":"TripAdvisor.co.uk","scraper":{"checkout_date":{"regex":"checkOut\".+?value=\"(.+?\\/.+?\\/.{4})\""},"location":{"regex":"\"location fl\">[\\S\\s]+?(.+?)<\\/div>[\\S\\s]+?<\\/div>[\\S\\s]<\\/li>[\\S\\s]<\\/ul>"},"hotel":{"regex":"h1 id=\"HEADING\" property=\"v:name\">[\\S\\s](.+?)<"},"checkin_date":{"regex":"checkIn\".+?value=\"(.+?\\/.+?\\/.{4})\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(book\\.)?bestwestern\\.com","name":"BestWestern.com","scraper":{"checkout_date":{"form_field":"input[name=\"depDate\"]:first"},"location":{"form_field":"#city"},"checkin_date":{"form_field":"input[name=\"arrDate\"]:first"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotwire\\.com","name":"Hotwire.com","scraper":{"checkout_date":{"form_field":"input[name=\"endDate\"]:first"},"location":{"form_field":"#hotelResultsFareFinder-destCity, #location"},"checkin_date":{"form_field":"input[name=\"startDate\"]:first"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
 var c = $ih("#siteSelected-button").children().first().attr('class'); if(c == 'label uk'){callback('%d/%m/%y');}else{callback('%m/%d/%y');}
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotelplanner\\.com","name":"HotelPlanner.com","scraper":{"checkout_date":{"form_field":"#CheckOutDate1"},"location":{"form_field":"#txtCity1"},"checkin_date":{"form_field":"#CheckInDate1"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?easytobook\\.com","name":"EasyToBook.com","scraper":{"checkout_date":{"regex":"departure: \"(.+?)\""},"location":{"form_field":"#location_name"},"hotel":{"regex":"class=\"hotelHeading\">(.+?)<"},"checkin_date":{"regex":"arrival: \"(.+?)\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d-%m-%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?onhotels\\.com","name":"OnHotels.com","scraper":{"checkout_date":{"form_field":"#CheckoutDateText"},"location":{"regex":"class=\"bold fs14 pl07\">(.+?)[,|-]"},"checkin_date":{"form_field":"#CheckinDateText"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %d %B %Y')
}
}},"category":"hotel","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?booking\\.de","name":"Booking.de","scraper":{"checkout_date":{"regex":"checkout_date = '(.{4}-.{2}-.{2})'"},"location":{"form_field":"#destination"},"hotel":{"form_field":"#hp_hotel_name"},"checkin_date":{"regex":"checkin_date = '(.{4}-.{2}-.{2})'"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?getaroom\\.com","name":"GetARoom.com","scraper":{"checkout_date":{"form_field":"#check_out"},"location":{"form_field":"#destination, .locality"},"hotel":{"regex":"class=\"fn org url faux-link\">(.+?)<"},"checkin_date":{"form_field":"#check_in"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotels\\.com\\/(hotel/|search.do?destination=)","name":"Hotels.com","scraper":{"checkout_date":{"form_field":"input[name=\"departureDate\"]:first, #search_checkOut"},"location":{"form_field":"input[name=\"destinationName\"]:first, input[name=\"hotelCity\"]:first, #search_destination"},"hotel":{"form_field":"input[name=\"hotelName\"]:first"},"checkin_date":{"form_field":"input[name=\"arrivalDate\"]:first, #search_checkIn"},"js":function(html, $ih, callback, scrapeField) {
var res = {};

var c = $ih(".flags").children().first().attr('class');
if(c == 'GBR')
{res.date_parsing_recipe = '%d\/%m\/%Y';}
else
{res.date_parsing_recipe = '%m\/%d\/%y';}

callback(res);

}
},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?cheapostay\\.com","name":"Cheapostay.com","scraper":{"checkout_date":{"form_field":"#txtCheckOutDate"},"location":{"form_field":"#ctl07_ctl00_ctl00_HotelUIMidScreen_FTop_txtHotelDestination"},"hotel":{"regex":"class=\"fn\">(.+?)<"},"checkin_date":{"form_field":"#txtCheckInDate,"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?travelrepublic\\.co.uk","name":"TravelRebublic.co.uk","scraper":{"checkout_date":{"js":function(html, $ih, callback, scrapeField) {
var match = $("#searchHeader").text(); if(match){ var date = match.match(/- \d{1,2} .+? \d{4}/);var dateNew = date[0].match(/\d{1,2} .+? \d{4}/); callback(dateNew[0])}else{callback();}
}
},"location":{"regex":"earchItinerary\">(.+?),"},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
var match = $("#searchHeader").text(); if(match){ var date = match.match(/\d{1,2} .+? \d{4}/);callback(date[0])}else{callback();}
}
},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %B %Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?alpharooms\\.com","name":"AlphaRooms.com","scraper":{"checkout_date":{"form_field":"#ctl00_MainContent_FloatingHotelSearchControl1_UsrCtlFlightHotelSearch_UsrCtlHotelCalenderControl_CheckOutPicker_TextBox"},"location":{"form_field":"#ctl00_MainContent_SearchResultsControl1_parentDestinationLinkButton"},"hotel":{"regex":"h1 class=\"fn org\">(.+?)<"},"checkin_date":{"form_field":"#ctl00_MainContent_FloatingHotelSearchControl1_UsrCtlFlightHotelSearch_UsrCtlHotelCalenderControl_CheckInPicker_TextBox"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?kayak\\.co.uk","name":"Kayak.co.uk","scraper":{"checkout_date":{"form_field":"#inline_checkout_date"},"location":{"form_field":"#inlinewherebox"},"checkin_date":{"form_field":"#inline_checkin_date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(.+?\\.)?hrs\\.com","name":"HRS.com","scraper":{"checkout_date":{"form_field":"#end_stayPeriod"},"location":{"form_field":"#destiny"},"checkin_date":{"form_field":"#start_stayPeriod"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?kayak\\.com","name":"Kayak.com","scraper":{"checkout_date":{"form_field":"#inline_checkout_date"},"location":{"form_field":"#inlinewherebox"},"checkin_date":{"form_field":"#inline_checkin_date"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?agoda\\.com","name":"Agoda.com","scraper":{"checkout_date":{"regex":"var checkedout = '(.+?)'"},"location":{"regex":"Hotels found in (.+?),"},"checkin_date":{"regex":"var checkedin = '(.+?)'"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?choicehotels\\.com","name":"ChoiceHotels.com","scraper":{"checkout_date":{"form_field":"input[name=\"departureDate\"]:first"},"location":{"form_field":"input[name=\"placename\"]:first"},"checkin_date":{"form_field":"input[name=\"arrivalDate\"]:first"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?otel\\.com","name":"Otel.com","scraper":{"checkout_date":{"regex":"s.prop5=\"(../../....)\""},"location":{"form_field":"#destination"},"hotel":{"regex":"<h1>(.+?)<"},"checkin_date":{"form_field":"#checkInDate"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?travelzoo\\.com","name":"TravelZoo.com","scraper":{"checkout_date":{"form_field":"#returnDate"},"location":{"form_field":"#txtDestination"},"checkin_date":{"form_field":"#departDate"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?splendia\\.com","name":"Splendia.com","scraper":{"checkout_date":{"form_field":"#label_dateend"},"location":{"form_field":"#searchInput"},"checkin_date":{"form_field":"#label_datestart"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d-%b-%y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?travelocity\\.com","name":"Travelocity.com","scraper":{"checkout_date":{"regex":"returningDate.+?(../../....)"},"location":{"regex":"\"city\":\"(.+?)\""},"hotel":{"regex":"([a-zA-Z0-9].+?) - .+? Hotel Information - Travelocity<"},"checkin_date":{"regex":"leavingDate.+?(../../....)"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?expedia\\.com\\/(Hotel-Search|.+?Hotel-Information)","name":"Expedia.com","scraper":{"checkout_date":{"form_field":"#inpEndDate, #hot-check-out, #hot-info-check-out"},"location":{"form_field":"#inpCity, #hot-sear-near, #hot-info-sear-near, #inpSearchNear"},"hotel":{"js":function(html, $ih, callback, scrapeField) {
 callback($ih('meta[property="og:title"]').attr('content'))
}
},"checkin_date":{"form_field":"#inpStartDate, #hot-check-in, #hot-info-check-in"},"rooms":{"form_field":"#hot-info-num-rooms, #inpRoom"},"adults":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumAdultsInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
},"children":{"js":function(html, $ih, callback, scrapeField) {
var s = 0; $ih('select[name="hotelSearchWizard_inpNumChildrenInRoom"]').each(function(i,e) {s += parseInt($ih(e).val())}); callback(s)
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?marriott\\.com","name":"Marriott.com","scraper":{"checkout_date":{"form_field":"#toDate"},"location":{"form_field":"#city-airport"},"checkin_date":{"form_field":"#fromDate"}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?olotels\\.com","name":"Olotels.com","scraper":{"checkout_date":{"regex":"to <span class=\"txt-heading\">(.+?)<"},"location":{"regex":"1 style=\"margin-bottom: 3px;\" class=\"left\">(.+?)<"},"checkin_date":{"regex":"From <span class=\"txt-heading\">(.+?)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hotelbookers\\.com","name":"HotelBookers.com","scraper":{"checkout_date":{"url_param":"checkout"},"location":{"url_param":"market"},"checkin_date":{"url_param":"checkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ratestogo\\.com","name":"RatesToGo.com","scraper":{"checkout_date":{"url_param":"hotel.chkout"},"location":{"url_param":"hotel.keyword.key"},"hotel":{"regex":"h1>(.+?)<"},"checkin_date":{"url_param":"hotel.chkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?tripadvisor\\.com\\/Hotel_Review","name":"TripAdvisor.com","scraper":{"checkout_date":{"regex":"per\" class=\" sprite-calendar-ylw text   popEnabled checkOut  no_cpu \" type=\"text\" value=\"(.+?\\/.+?\\/.{4})\""},"location":{"regex":"<title>.+? \\((.+?)\\)"},"hotel":{"regex":"h1 id=\"HEADING\" property=\"v:name\">[\\S\\s]+?(.+?)$"},"checkin_date":{"regex":"per\" class=\" sprite-calendar-ylw text   popEnabled checkIn  no_cpu \" type=\"text\" value=\"(.+?\\/.+?\\/.{4})\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"hotel","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ebookers\\.com","name":"Ebookers.com","scraper":{"checkout_date":{"url_param":"hotel.chkout"},"location":{"url_param":"hotel.keyword.key"},"hotel":{"regex":"h1>(.+?)<"},"checkin_date":{"url_param":"hotel.chkin"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%y')
}
}},"category":"hotel","style_override":null}]};
com.forward.invisiblehand.rentals.scrapers = {"regions":{"uk":{"name":"United Kingdom","code":"UK","server":"uk.lb.invisiblehand.co.uk","currency":{"name":"GBP","symbol":"\u00a3"}},"de":{"name":"Germany","code":"DE","server":"de.lb.invisiblehand.co.uk","currency":{"name":"EUR","symbol":"\u20ac"}},"ca":{"name":"Canada","code":"CA","server":"ca.lb.invisiblehand.co.uk","currency":{"name":"CAD","symbol":"$"}},"us":{"name":"United States","code":"US","server":"us.lb.invisiblehand.co.uk","currency":{"name":"USD","symbol":"$"}}},"operators":[{"region":"UK","regex":"^https?://(www\\.)?alamo\\.co.uk","name":"Alamo.co.uk","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
callback("10/10/2010")
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?budget\\.co.uk","name":"Budget.co.uk","scraper":{"sipp":{"element_text":".carType:first"},"checkout_date":{"regex":",doDate=(..........)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"\"returnStation\">(.+?)<"},"checkin_date":{"regex":",puDate=(..........)"},"scrapped_price":{"element_text":".carPrice:first"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?easycar\\.com","name":"Easycar.com","scraper":{"checkout_date":{"form_field":"#txtEndDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"element_text":"#lblPickupLoc:first"},"checkin_date":{"form_field":"#txtStartDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?carhire3000\\.com","name":"Carhire3000.com","scraper":{"checkout_date":{"element_text":"#datepicker2"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"url_param":"city"},"checkin_date":{"element_text":"#datepicker"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"element_text":".result_now:first"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?bookryanair\\.com\\/SkySales\\/booking\\.aspx\\?culture=en-gb\\&lc=en-gb\\#CarHire","name":"BookRyanAir.com","scraper":{"checkout_date":{"element_text":"#ChangeMainDiv div fieldset:eq(1) div:eq(0) div:eq(1) time"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"element_text":".Station:first"},"checkin_date":{"element_text":"#ChangeMainDiv div fieldset:eq(0) div:eq(0) div:eq(1) time"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %d %b %Y %H:%M')
}
},"scrapped_price":{"element_text":".prc:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?dollar\\.com","name":"Dollar.com","scraper":{"checkout_date":{"regex":"ctl07_ctl01_ReservationStatusColumnLayout_ReturnAirportDateLabel\">(... ...\\. .., ....) ."},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"element_text":"#ctl07_ctl01_ReservationStatusColumnLayout_PickupAirportLabel"},"checkin_date":{"regex":"ctl07_ctl01_ReservationStatusColumnLayout_PickupAirportDateLabel\">(... ...\\. .., ....) ."},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a %b. %d, %Y')
}
},"currency":{"element_text":".vehGuideRatePeriod:first, .rdAmount:last"},"scrapped_price":{"element_text":"#ctl07_ctl03_AllCarsColumnLayout_VehicleList_VehicleRatesRepeater_RateTotalDescLabel_0"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?avis\\.co.uk","name":"Avis.co.uk","scraper":{"sipp":{"form_field":"#main_fleetControl_fleetRepeater_vehicleModelControl_0_vehicleDisplayName_0_vehicleTypeLabel_0"},"checkout_date":{"url_param":"ReturnDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"url_param":"PickupCode"},"checkin_date":{"url_param":"PickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
},"scrapped_price":{"form_field":"#main_fleetControl_fleetRepeater_vehicleModelControl_0_priceLabel_0"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?alamo\\.com","name":"Alamo.com","scraper":{"checkout_date":{"form_field":".puLocdetails dd  br:eq(1)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"form_field":".doLocdetails dd address br:eq(1)"},"checkin_date":{"form_field":".puLocdetails dd  br:eq(1)"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?hertz\\.com\\/rentacar\\/reservation","name":"Hertz.com","scraper":{"sipp":{"regex":"class=\"vehGuideCarSippCode gblRemoved\">(.+?)<"},"checkout_date":{"regex":"rItnDropoffLocSec[\\S\\s]+?(..., \\d\\d ..., \\d\\d\\d\\d)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"checkin_date":{"regex":"rItnPickupLocSec[\\S\\s]+?(..., \\d\\d ..., \\d\\d\\d\\d)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %d %b, %Y')
}
},"currency":{"element_text":".vehGuideRatePeriod:first, .rdAmount:last"},"scrapped_price":{"element_text":".vehGuideRate:first, .rdAmount:last"},"js":function(html, $ih, callback, scrapeField) {

var res = {};

var a = $ih("#rItnPickupLocSec").html().match(/mp_trans_disable_end -->[\S\s]+?\s(\S.+)/);

if (a.length > 1) res.location = a[1];

callback(res);
}
},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?autoeurope\\.com\\/car","name":"AutoEurope.com","scraper":{"sipp":{"form_field":"input[name=\"sipp\"]:first"},"checkout_date":{"form_field":".step2_refineItin table tbody tr td:eq(5)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"element_text":".refine_itinTblTopRow td:first"},"checkin_date":{"form_field":".step2_refineItin table tbody tr td:eq(4)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y')
}
},"scrapped_price":{"form_field":".cd_pkgCost .cd_pkgPrice:eq(1)"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?atlaschoice\\.com","name":"AtlasChoice.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
callback("10/10/2010")
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(shopping\\.)?usairways\\.com\\/Cars\\/","name":"UsAirways.com","scraper":{"checkout_date":{"form_field":"#ctl00_phMain_ModifySearch1_ModifySearchModule_ctl00_txtCarsDropoffDate_SelectedDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#ctl00_phMain_ModifySearch1_ModifySearchModule_ctl00_txtCarsPickup_textbox"},"checkin_date":{"form_field":"#ctl00_phMain_ModifySearch1_ModifySearchModule_ctl00_txtCarsPickupDate_SelectedDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?sixt\\.co.uk","name":"Sixt.co.uk","scraper":{"checkout_date":{"element_text":"#sx-js-res-ret-date"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"element_text":".sx-gc-txt:first"},"checkin_date":{"element_text":"#sx-js-res-pu-date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d.%m.%Y')
}
},"scrapped_price":{"element_text":"..sx-res-offerlist-offerprice:eq(0) p strong"},"currency":{"element_text":".sx-res-offerlist-offerprice:eq(0) p strong"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?nationalcar\\.co\\.uk","name":"NationalCar.co.uk","scraper":{"checkout_date":{"regex":"Drop off location: (..? ... ....)","rescrape_for":5},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"Pick up location: .+? .+?:.+?  - (.+?)<","rescrape_for":5},"checkin_date":{"regex":"Pick up location: (..? ... ....)","rescrape_for":5},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
},"scrapped_price":{"form_field":"input[class=\"cssAlamoPrice\"]:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?acerentacar\\.com\\/resselect\\.aspx","name":"AceRentacar.com","scraper":{"checkout_date":{"regex":"Return Date\\/Time<\\/b>[\\S\\s]+?<td style=\"padding-left: 2px\">[\\S\\s]+?<span style=\"font-size: small\">[\\S\\s]+?(... ..?, ....) \\@"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"Pickup Location<\\/b>[\\S\\s]+?<td style=\"padding-left: 2px\">[\\S\\s]+?<span style=\"font-size: small\">[\\S\\s]+?(\\w.+?)<"},"checkin_date":{"regex":"Pickup Date\\/Time<\\/b>[\\S\\s]+?<td style=\"padding-left: 2px\">[\\S\\s]+?<span style=\"font-size: small\">[\\S\\s]+?(... ..?, ....) \\@"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y')
}
},"scrapped_price":{"regex":"<td style=\"padding-right: 2px; color: #0b306c\" align=\"right\">[\\S\\s]+?<strong>[\\S\\s]+?\\D(\\d.+?\\.?\\d?)\\s"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?holidaycars\\.com\\/select_car","name":"HolidayCars.com","scraper":{"sipp":{"form_field":"input[name=\"sipp\"]:first"},"checkout_date":{"form_field":"#dropoffdate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"element_text":"#location_ac"},"checkin_date":{"form_field":"#pickupdate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
},"scrapped_price":{"js":function(html, $ih, callback, scrapeField) {
var price = $ih('#deprijs').text().match(/[\d].+/); if(price){callback(price[0]);}else {callback('');}
}
}},"category":"rentals","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?budget\\.de","name":"Budget.de","scraper":{"checkout_date":{"regex":",doDate=(..........)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('DE')
}
},"location":{"regex":"PULoc=(.+?)&"},"checkin_date":{"regex":",puDate=(..........)"},"scrapped_price":{"element_text":".carPrice:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?paylesscar\\.com\\/reservation","name":"PaylessCar.com","scraper":{"checkout_date":{"form_field":"#dynaSearchForm_rd"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#dynaSearchForm_locDesc"},"checkin_date":{"form_field":"#dynaSearchForm_pd"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"form_field":"input[class=\"price_total\"]:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?expedia\\.com\\/pub\\/agent\\.dll\\?qscr=cars","name":"Expedia.com","scraper":{"checkout_date":{"url_param":"date2"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"url_param":"locn"},"checkin_date":{"url_param":"date1"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"regex":"from .(.+?) Base price"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(trip\\.)?delta\\.com\\/cars\\/","name":"Delta.com","scraper":{"checkout_date":{"form_field":"#end_date"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#start_location"},"checkin_date":{"form_field":"#start_date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?economyrentacar\\.com\\/reservation\\.aspx","name":"EconomyRentacar.com","scraper":{"checkout_date":{"form_field":"#ctl00_ContentPlaceHolder1_txtDropDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"id=\"ctl00_ContentPlaceHolder1_txtPick\">. (.+?)[<|(]","rescrape_for":5},"checkin_date":{"form_field":"#ctl00_ContentPlaceHolder1_txtPickDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"form_field":"#ctl00_ContentPlaceHolder1_dtlCars_ctl00_lbl_c3"}},"category":"rentals","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?easycar\\.de","name":"Easycar.de","scraper":{"checkout_date":{"form_field":"#txtEndDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('DE')
}
},"location":{"element_text":"#lblPickupLoc:first"},"checkin_date":{"form_field":"#txtStartDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?auto-europe\\.co\\.uk\\/car","name":"AutoEurope.co.uk","scraper":{"sipp":{"form_field":".carClass:eq(0)"},"checkout_date":{"form_field":".step2_refineItin table tbody tr td:eq(5)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"element_text":".refine_itinTblTopRow td:first"},"checkin_date":{"form_field":".step2_refineItin table tbody tr td:eq(4)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
},"scrapped_price":{"form_field":".cd_pkgCost .cd_pkgPrice:eq(1)"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?holidayautos\\.co\\.uk\\/cgi-bin","name":"HolidayAutos.co.uk","scraper":{"sipp":{"regex":"sipp_code:\"(.+?)\""},"checkout_date":{"regex":"span class=\"label-width left font-bold\">to:<\\/span> (.+?)&"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"regex":"<span class=\"label-width left font-bold\">pick up:</span> (.+?) <"},"checkin_date":{"regex":"span class=\"label-width left font-bold\">from:<\\/span> (.+?)&"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"regex":"online_price:\"(.+?)\""}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?expedia\\.co\\.uk\\/carsearch?","name":"Expedia.co.uk","scraper":{"checkout_date":{"url_param":"date2"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"url_param":"locn"},"checkin_date":{"url_param":"date1"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"regex":"from \\\u00a3(.+?) "}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?enterprise\\.co.uk","name":"Enterprise.co.uk","scraper":{"sipp":{"element_text":".normTextBold[3]"},"checkout_date":{"regex":"Return<.strong>[\\S\\s]+?(..? ... \\d{4})"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"regex":">Pick Up Location<\\/strong>[\\S\\s]+?<br>[\\S\\s]+?([A-Z]+?),.+?<br>"},"checkin_date":{"regex":"Pick Up<.strong>[\\S\\s]+?(..? ... \\d{4})"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
},"scrapped_price":{"element_text":"#tdCCTotalPrice, #tdCDTotalCharges"},"currency":{"element_text":"#tdCCTotalPrice, #tdCDTotalCharges"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?europcar\\.com","name":"Europcar.com","scraper":{"checkout_date":{"form_field":"#to"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"checkin_date":{"form_field":"#from"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":".total:eq(1)","rescrape_for":10},"currency":{"element_text":".total:eq(1)","rescrape_for":10},"js":function(html, $ih, callback, scrapeField) {

var res = {};

var a = $ih(".location:first").text().match(/(\w.+)/);

res.location = a[1];

callback(res);
}
},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?travelocity\\.com\\/Car-Rental","name":"Travelocity.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"checkin_date":{"js":function(html, $ih, callback, scrapeField) {
callback("10/10/2010")
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(cars\\.)?united\\.com","name":"United.com","scraper":{"checkout_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spDropoffDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupLocation"},"checkin_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a., %b. %d, %Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?continental\\.com\\/web\\/en-US\\/apps\\/booking\\/car\\/searchResults\\.aspx","name":"Continental.com","scraper":{"checkout_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spDropoffDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"js":function(html, $ih, callback, scrapeField) {
var price = $(".featuredCar:first .Notes").html().match(/\$\d{1,5}/); if(price){callback(price[0]);}else{callback()}
}
},"location":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupLocation"},"checkin_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a., %b. %d, %Y')
}
}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?rentalcars\\.com\\/","name":"RentalCars.com","scraper":{"sipp":{"element_text":".result_class:first"},"checkout_date":{"element_text":"#datepicker2"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"form_field":"#city"},"checkin_date":{"element_text":"#datepicker"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"element_text":".result_now:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?orbitz.com\\/App\\/ViewCarResults","name":"Orbitz.com","scraper":{"checkout_date":{"form_field":"#carchgEndDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"class=\"searchedLocation\">(.+?)<"},"checkin_date":{"form_field":"#carchgStartDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?budget\\.com","name":"Budget.com","scraper":{"sipp":{"element_text":".carType:first"},"checkout_date":{"regex":",doDate=(..........)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"PUL=(.+?),"},"checkin_date":{"regex":",puDate=(..........)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"regex":"id=\"totalRate\">[\\s]*\\$[\\s]*(\\d+\\.\\d+)"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?hertz4ryanair\\.com","name":"Hertz4RyanAir.com","scraper":{"sipp":{"regex":"class=\"vehGuideCarSippCode gblRemoved\">(.+?)<"},"checkout_date":{"regex":"<dt>Drop off Date\\/Time - <\\/dt>[\\S\\s].+<dd>(\\d.+?)\\/"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"regex":"<dt>Pick up Location - <\\/dt>[\\S\\s].+<dd>(\\D.+?)<"},"checkin_date":{"regex":"<dt>Pick up Date\\/Time - <\\/dt>[\\S\\s].+<dd>(\\d.+?)\\/"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y ')
}
},"scrapped_price":{"element_text":".selectPrice:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?sixt\\.com","name":"Sixt.com","scraper":{"checkout_date":{"element_text":"#sx-js-res-ret-date"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"element_text":".sx-gc-txt:first"},"checkin_date":{"element_text":"#sx-js-res-pu-date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":"..sx-res-offerlist-offerprice:eq(0) p strong"},"currency":{"element_text":".sx-res-offerlist-offerprice:eq(0) p strong"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(.+?\\.)?budgetinternational\\.com","name":"BudgetInternational.com","scraper":{"checkout_date":{"form_field":"#ettDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"location":{"regex":"\"returnStation\">(.+?)<"},"checkin_date":{"form_field":"#etaDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"element_text":".price:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?foxrentacar\\.com","name":"FoxRentalCar.com","scraper":{"checkout_date":{"regex":"tagItinerary_LblDropOffDateTime\" class=\"small_text\">(..........)<"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"tagItinerary_LblPickupLoc\" class=\"small_text\">\\*(.+?)<"},"checkin_date":{"regex":"tagItinerary_LblPickupDateTime\" class=\"small_text\">(..........)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(travel.\\.)?priceline.com\\/rentalcars\\/","name":"Priceline.com","scraper":{"sipp":{"rescrape_for":2,"js":function(html, $ih, callback, scrapeField) {
var sipp = $ih('#partner1_1 .carInfo').text().match(/.([a-zA-Z]{4})/)[1]; if(sipp){callback(sipp);}else {callback('');}
}
},"checkout_date":{"form_field":"input[name=dropOffDate]:first","rescrape_for":2},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"input[name=pickUpSearchAirportString]:first","rescrape_for":2},"checkin_date":{"form_field":"input[name=pickUpDate]:first","rescrape_for":2},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":"#partner1_1 div:eq(0)","rescrape_for":2}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?carrentalexpress\\.com","name":"CarRentalExpress.com","scraper":{"checkout_date":{"form_field":"#date_dropoff"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"element_text":"#_pickup_poi"},"checkin_date":{"form_field":"#date_pickup"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?southwest\\.com\\/car-rentals\\/select-car\\.html","name":"Southwest.com","scraper":{"checkout_date":{"form_field":"#returnDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#pickUpLocationText"},"checkin_date":{"form_field":"#outboundDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":".vehGuideRate:first, .rdAmount:last"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?thrifty\\.com","name":"Thrifty.com","scraper":{"checkout_date":{"form_field":"#reservations_controls_reservationprogress_ascx1_DropOffDateTextLabel"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#reservations_controls_reservationprogress_ascx1_PickupLocationTextLabel"},"checkin_date":{"form_field":"#reservations_controls_reservationprogress_ascx1_PickupDateTextLabel"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y %I:%M %p')
}
}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(reservations\\.)?zoomrentals\\.com\\/rates?","name":"ZoomRentals.com","scraper":{"checkout_date":{"regex":"Dropoff Date:<\\/div>[\\S\\s]+?<div class=\"side_res_summary_data\">(... ..?, ....) "},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"Pickup Location:<\\/div>[\\S\\s]+?<div class=\"side_res_summary_data\">(.+?)<","rescrape_for":5},"checkin_date":{"regex":"Pickup Date:<\\/div>[\\S\\s]+?<div class=\"side_res_summary_data\">(... ..?, ....) "},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y')
}
},"scrapped_price":{"form_field":"input[class=\"rates_r2_rate\"]:first"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?enterprise\\.com","name":"Enterprise.com","scraper":{"sipp":{"element_text":".normTextBold:eq(3)"},"checkout_date":{"regex":"Return<.strong>[\\S\\s]+?(... ..?, ....)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":">Pick Up Location<\\/strong>[\\S\\s]+?<br>[\\S\\s]+?([A-Z]+?),.+?<br>"},"checkin_date":{"regex":"Pick Up<.strong>[\\S\\s]+?(... ..?, ....)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b %d, %Y')
}
},"scrapped_price":{"element_text":"#tdCCTotalPrice, #tdCDTotalCharges"},"currency":{"element_text":"#tdCCTotalPrice, #tdCDTotalCharges"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?avis\\.com","name":"Avis.com","scraper":{"sipp":{"element_text":".carClassText:first, #carClassInfoTxt"},"checkout_date":{"regex":"doDate=(../../....)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"input[name=\"resForm.pickUpLocation\"]:first"},"checkin_date":{"regex":"puDate=(.+?),"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":".payonlinenow:first, #estiTotal"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?carrentals\\.com","name":"CarRentals.com","scraper":{"sipp":{"element_text":".rateDetails:eq(0) ul li"},"checkout_date":{"form_field":"#dropoffDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"js":function(html, $ih, callback, scrapeField) {
var price = $('.ratePrice:eq(0) p strong').text(); if(price){callback(price);}else{callback('');}
}
},"location":{"form_field":"#location"},"checkin_date":{"form_field":"#pickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"scrapped_price":{"element_text":".ratePrice:eq(0) p strong"}},"category":"rentals","style_override":null},{"region":"US","regex":"e-zrentacar\\.com\\/results\\.asp","name":"Ezrentacar.com","scraper":{"checkout_date":{"regex":"Same As Pickup<br \\/>[\\S\\s]+?(...\\&nbsp;..?,&nbsp;....)\\&"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"target=\"_blank\"><span style=\"font-size:14px;\">(.+?)<","rescrape_for":5},"checkin_date":{"regex":"<\\/span><\\/a><br \\/>[\\S\\s]+?(...\\&nbsp;..?,&nbsp;....)\\&"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%b&nbsp;%d,&nbsp;%Y')
}
},"scrapped_price":{"form_field":"input[class=\"estTotal\"]:first"}},"category":"rentals","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?sixt\\.de","name":"Sixt.de","scraper":{"checkout_date":{"element_text":"#sx-js-res-ret-date"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('DE')
}
},"location":{"element_text":".sx-gc-txt:first"},"checkin_date":{"element_text":"#sx-js-res-pu-date"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d.%m.%Y')
}
},"scrapped_price":{"element_text":"..sx-res-offerlist-offerprice:eq(0) p strong"},"currency":{"element_text":".sx-res-offerlist-offerprice:eq(0) p strong"}},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?skyscanner\\.net\\/car-hire\\/","name":"SkyScanner.net","scraper":{"checkout_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spDropoffDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupLocation"},"checkin_date":{"form_field":"#ctl00_ContentInfo_AvailabilitySummary_spPickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a., %b. %d, %Y')
}
}},"category":"rentals","style_override":null},{"region":"DE","regex":"^https?://(www\\.)?avis\\.de","name":"Avis.de","scraper":{"checkout_date":{"url_param":"ReturnDate"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('DE')
}
},"location":{"url_param":"PickupCode"},"checkin_date":{"url_param":"PickupDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
},"scrapped_price":{"form_field":"#main_fleetControl_fleetRepeater_vehicleModelControl_0_priceLabel_0"}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?hertz\\.co.uk\\/rentacar\\/reservation\\/","name":"Hertz.co.uk","scraper":{"sipp":{"regex":"class=\"vehGuideCarSippCode gblRemoved\">(.+?)<"},"checkout_date":{"regex":"rItnDropoffLocSec\">[\\S\\s]+href.+[\\S\\s]+(..., .. ..., ....)"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"checkin_date":{"regex":"rItnPickupLocSec\">[\\S\\s]+?href.+?[\\S\\s]+?(..., .. ..., ....)"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %d %b, %Y')
}
},"currency":{"element_text":".vehGuideRatePeriod:first, .rdAmount:last"},"scrapped_price":{"element_text":".vehGuideRate:first, .rdAmount:last"},"js":function(html, $ih, callback, scrapeField) {

var res = {};

var a = $ih("#rItnPickupLocSec").text().match(/at ..:..[\S\s]+?(\S.+)/);

res.location = a[1];

callback(res);
}
},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?nationalcar\\.com","name":"NationalCar.com","scraper":{"checkout_date":{"regex":"n2=escape('(.+?)');","rescrape_for":20},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"regex":"Pickup:[\\S\\s]+?class=\"searchText\">(.+?)<","rescrape_for":20},"checkin_date":{"regex":"x6=escape('(.+?)');","rescrape_for":20},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%A, %B %d, %Y')
}
}},"category":"rentals","style_override":null},{"region":"UK","regex":"^https?://(www\\.)?europcar\\.co.uk","name":"Europcar.co.uk","scraper":{"checkout_date":{"form_field":"#to"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('UK')
}
},"checkin_date":{"form_field":"#from"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d/%m/%Y')
}
},"scrapped_price":{"wait_for":5,"element_text":".total:eq(1)"},"currency":{"wait_for":5,"element_text":".total:eq(1)"},"js":function(html, $ih, callback, scrapeField) {

var res = {};

var a = $ih(".location:first").text().match(/(\w.+)/);

res.location = a[1];

callback(res);
}
},"category":"rentals","style_override":null},{"region":"US","regex":"^https?://(www\\.)?kayak.com\\/.?cars","name":"Kayak.com","scraper":{"checkout_date":{"form_field":"input[id=inline_dropoff_date]:first"},"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"location":{"form_field":"input[id=inlinepickuplocation]:first"},"checkin_date":{"form_field":"input[id=inline_pickup_date]:first"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"rentals","style_override":null}]};
com.forward.invisiblehand.tickets.scrapers = {"regions":{"uk":{"name":"United Kingdom","code":"UK","server":"uk.lb.invisiblehand.co.uk","currency":{"name":"GBP","symbol":"\u00a3"}},"de":{"name":"Germany","code":"DE","server":"de.lb.invisiblehand.co.uk","currency":{"name":"EUR","symbol":"\u20ac"}},"ca":{"name":"Canada","code":"CA","server":"ca.lb.invisiblehand.co.uk","currency":{"name":"CAD","symbol":"$"}},"us":{"name":"United States","code":"US","server":"us.lb.invisiblehand.co.uk","currency":{"name":"USD","symbol":"$"}}},"operators":[{"region":"US","regex":"^https?://(www\\.)?ebay\\.com","name":"Ebay.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"regex":"itemprop=\"price\" .+?>US \\$(.+?)<","rescrape_for":5},"date_parsing_recipe":{"rescrape_for":5,"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
},"js":function(html, $ih, callback, scrapeField) {
var isTicketSearch = function(query) {
  var match = query.match(/(tickets)/i);
  return !!match;
}

var string_match = html.match(/h1 itemprop="name" class="vi-is1-titleH1">(.+?)</);
var query = string_match[1];
if (!isTicketSearch(query)) return;

var res = {};

var match = html.match(/Event Name: .+? width="50.0%">(.+?)</);
if (match) res.artist = match[1];

if (res.artist == null){
var match = query.match(/\d (.+) tickets/i);
if (match) res.artist = match[1];}

if (res.artist == null){
var match = query.match(/(.+) tickets/i);
if (match) res.artist = match[1];}

var match = html.match(/Event Date: .+? width="50.0%">(.+?)</);
if (match) res.date = match[1];

if(res.date == null){
var match = query.match(/(\d+\/\d+\/\d+)/);
if (match) res.date = match[1];}

if(res.date == null){
var match = query.match(/\s(\d+\/\d+)\s/);
var d = new Date();
var year = d.getFullYear();
if (match) res.date = match[1] + "/" + year.toString();}


callback(res);
}
},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?razorgator\\.com","name":"Razorgator.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"regex":"Price Range <span style=\"font-weight:bold\">\\$(.+?) -"},"artist":{"regex":"var Event = \"(.+?)\""},"date":{"regex":"var Date = \"(.+?) "},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ticketsnow\\.com","name":"TicketsNow.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"regex":"id=\"ucTicketList_lblMinPrice\">\\$(.+?)<"},"artist":{"regex":"dtmTag.dtmc_artist_team = \"(.+?)\""},"date":{"regex":"dtmTag.dtmc_event_date = \"(.+?)\""},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%Y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ticketnetwork\\.com","name":"TicketsNetwork.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".tixPriceLine:first"},"artist":{"regex":"id=\"results_page_title\">(.+?)<"},"date":{"regex":"id=\"evtDateDtls_rt_hdr\">(.+?) at <"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%A, %B %d, %Y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?seatwave\\.com","name":"SeatWave.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".price[1]"},"artist":{"regex":"<h1>(.+?)<"},"date":{"regex":"</a></li><li>(.. ... ....)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%d %b %Y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ticketmaster\\.com","name":"TicketMaster.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"artist":{"regex":"link=edp_Artist_Name\">(.+?)<"},"date":{"regex":"id=\"artist_event_date\">(.+?)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %b %d, %Y %I:%M %p')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?viagogo\\.com","name":"Viagogo.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"regex":"\"pleaseWaitElementID\":\"ctl00_centerContent_loadingControl\",\"records\":\\[\\[(.+?),"},"artist":{"regex":"class=\"go_widget_accordian_header\"><span>(.+?)<"},"date":{"regex":"itemprop=\"startDate\" content=\"(.+?)T"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?vividseats\\.com","name":"VividSeats.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".ticketColumn5:first"},"artist":{"regex":"id=\"title\"><h3>(.+?)<"},"date":{"regex":"id=\"date\">(.+?)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a. %b %d, %Y %I:%M %p')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ticketcity\\.com","name":"TicketCity.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".tdPrice:first"},"artist":{"form_field":"#EventHeading"},"date":{"regex":"event_date\"><span>(.+?;\\d{4})"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%A,&nbsp;%B&nbsp;%d,&nbsp;%Y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?stubhub\\.com","name":"TicketsStubHub.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"artist":{"regex":"id=\"eventSummary\"><div class=\"p1\"><div class=\"eventTitle\">(.+?)<"},"date":{"regex":"id=\"h1title\">.+? (....-.+-.+)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%Y-%m-%d')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(purchase\\.)?tickets\\.com","name":"Tickets.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".ticketColumn5:first"},"artist":{"regex":"class=\"perfdesctext\">(.+?)<"},"date":{"regex":"class=\"perfdaytext\">.+?, (.+?),"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%m/%d/%y')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?ticketliquidator\\.com","name":"TicketLiquidator.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".ssc_lrPri:first"},"artist":{"element_text":"#eventName"},"date":{"element_text":"#displayDate"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %b %d, %Y at %I:%M %p')
}
}},"category":"tickets","style_override":null},{"region":"US","regex":"^https?://(www\\.)?bigstub\\.com","name":"BigStub.com","scraper":{"region":{"js":function(html, $ih, callback, scrapeField) {
callback('US')
}
},"price":{"element_text":".ssc_lrPri:first"},"artist":{"regex":".aspx\" title=\"(.+?) Tickets\">"},"date":{"regex":"tn_results_header_datetime\">(.+?)<"},"date_parsing_recipe":{"js":function(html, $ih, callback, scrapeField) {
callback('%a, %b %d %Y')
}
}},"category":"tickets","style_override":null}]};
